#define DIALOGON 0
/* #define MSGCHECK 0 */

#include <stdio.h>
#include <string.h>
#include "Xiomgr.h"

extern int system(char *);

/*
 *  Constructor insures msg files empty.
 */
 msg::msg() {
   FILE *fp;
   nrBufUsed = 0;
   if ((fp = fopen("IO.output", "w")) == NULL)
     fprintf(stderr, "Couldn't initialize output message file\n");
   else {
     fprintf(fp, "%d\n", 0);
     fclose(fp);
   }
   if ((fp = fopen("IO.input", "w")) == NULL)
     fprintf(stderr, "Couldn't initialize input message file\n");
   else {
     fprintf(fp, "%d\n", 0);
     fclose(fp);
   }
 }

/*
 *  Destructor empties files before quitting.
 */
 msg::~msg() {
   FILE *fp;
   if ((fp = fopen("IO.output", "w")) != NULL) {
     fprintf(fp, "%d\n", 0);
     fclose(fp);
   }
   if ((fp = fopen("IO.input", "w")) != NULL) {
     fprintf(fp, "%d\n", 0);
     fclose(fp);
   }
 }

/*
 *  Zero message buffer and put one line in it.
 */
void msg::newMsgBuf(int m1, int m2, msgstr msg) {
  sprintf(outbuf[0], "%d %d %s\n", m1, m2, msg);
  nrBufUsed = 1;
}

/*
 *  Add one line to current msg buffer.  Return nr in buffer; 0 if full.
 */
int msg::addtoMsgBuf(int m1, int m2, msgstr msg) {
  if (nrBufUsed == NRBUFFERS) return 0;
  sprintf(outbuf[nrBufUsed++], "%d %d %s\n", m1, m2, msg);
  return nrBufUsed;
}

/*
 *  Send a one line Msg.  Return 1 if sent, 0 if unable to send.
 */
int msg::sendOnelineMsg(int m1, int m2, msgstr msg) {
  FILE *fp;
  int nrInFile;
  if ((fp = fopen("IO.output", "r+")) == NULL) return 0;
  fscanf(fp, "%d ", &nrInFile);
  if (nrInFile > 0) {
    fclose(fp);
    return -1;
  }
  fseek(fp, 0, 0);
  fprintf(fp, "%d\n", 1);
  fprintf(fp, "%d %d %s\n", m1, m2, msg);
  fclose(fp);
  nrBufUsed = 0;
#ifdef MSGCHECK
  printf("sent: %d %d %s\n", m1, m2, msg);
#endif
  return 1;
}

/*
 *  Send out whatever is in the msg buffer. Return nr sent if done, 0 if not.
 */
int msg::sendMsgBuf() {
  FILE *fp;
  int i, nrInFile;
  if ((fp = fopen("IO.output", "r+")) == NULL) return 0;  /* couldn't open */
  fscanf(fp, "%d ", &nrInFile);
  if (nrInFile > 0) {
    fclose(fp);
    return -1;    /* can't write - not read yet */
  }
  fseek(fp, 0, 0);
  fprintf(fp, "%d\n", nrBufUsed);
  for (i = 0; i < nrBufUsed; i++)
    fprintf(fp, "%s", outbuf[i]);
  fclose(fp);
  nrBufUsed = 0;
  return i;
}

/*
 *  Read incoming messages.  Return 0 if cannot open file, -1 if file empty,
 *    or nr messages read.
 */
int msg::recvMsgBuf(msgbuf inbuf) {
  FILE *fp;
  int i, nrtoread;
  msgstr msg;
  if ((fp = fopen("IO.input", "r+")) == NULL) {
    return 0;  /* couldn't open */
  }
  fscanf(fp, "%d ", &nrtoread);
  if (nrtoread == 0 || nrtoread > 10) {
    fclose(fp);
    return -1;
  }
  for (i = 0; i < nrtoread; i++) {
    fgets(inbuf[i], sizeof(commstr), fp);
#ifdef MSGCHECK    
    printf("recd: %s", inbuf[i]);
#endif
  }
  fseek(fp, 0, 0);
  fprintf(fp, "%d\n", 0);
  fclose(fp);
  return nrtoread;
}

/*
 *  Break msg line up into int,int,str.  Needed because str may have blanks.
 */
void msg::decipherMsg(commstr buf, int& m1, int& m2, msgstr msg) {
  char *ptr;
  sscanf(buf, "%d%d", &m1, &m2);  /* get the 2 int's */
  ptr = strchr(buf, ' ') + 1;   /* point to char after 1st blank */
  ptr = strchr(ptr, ' ') + 1;   /* point to char after 2nd blank */
  strcpy(msg, ptr);             /* copy the remainder to string */
  msg[strlen(msg)-1] = '\0';    /* dump the \n at end */
}
  
 iomgr::iomgr() {
   nextToRead = 0;
   nrInWaiting = 0;
   logfileon = 0;
   dialogBoxOpen = 0;
}

void iomgr::logfileOnOff(mstring inputstr) {
  if (logfileon) {
    logfileon = 0;
    fclose(logfp);
  }
  else {
    if ((logfp = fopen(inputstr, "w")) == NULL)
	printf("Unable to open logfile %s\n", inputstr);
    else 
      logfileon = 1;
  }
}

int iomgr::getnextinput(int& m2choice, msgstr inputdata) {
  int m1choice = 0;
  int nrRead;
  m2choice = 0;
  if (nextToRead < nrInWaiting) {
    msgMgr.decipherMsg(inbuf[nextToRead++], m1choice, m2choice, inputdata);
    return m1choice;
  }
  nextToRead = nrInWaiting = 0;
  nrRead = msgMgr.recvMsgBuf(inbuf);
  if (nrRead > 0) {
    msgMgr.decipherMsg(inbuf[nextToRead++], m1choice, m2choice, inputdata);
    nrInWaiting = nrRead;
    return m1choice;
  }
  return 0;
}

int iomgr::savefilefirst(mstring filename) {
  printf("%s not saved before loading...\n", filename);
  return 1;
}

#ifdef DIALOGON
/*
 *  Send msg to open Xwindow dialog box 'nrlines' long.
 */
void iomgr::openDialog(int nrlines){
  msgstr msg;
  if (dialogBoxOpen) return;
  if (nrlines < 0 || nrlines >24) nrlines = 10;
  sprintf(msg,"%d", nrlines);
  while(msgMgr.sendOnelineMsg(210, 1, msg) <= 0);
  dialogBoxOpen = 1;
}

/*
 *  Send msg to close the Dialog Xwindow.
 */
void iomgr::closeDialog() {
  msgstr msg;
  if(!dialogBoxOpen) return;
  msg[0] = '\0';
  while(msgMgr.sendOnelineMsg(210, 2, msg) <= 0);
  dialogBoxOpen = 0;
}

void iomgr::cancelDialogButton() {
  msgstr msg;
  if(!dialogBoxOpen) return;
  msg[0] = '\0';
  while(msgMgr.sendOnelineMsg(210, 4, msg) <= 0);
  dialogBoxOpen = 0;
}

/*
 *  Print a line. 'Where' controls where msg goes.
 *  NOPRESS = No 'Press key'; PRESS = 'Press key'; LOG = logfile only; NOLOG
 */
void iomgr::printline(int where, mstring data) {
  if (where != LOG) {
    if (!dialogBoxOpen) openDialog(10);
    while(msgMgr.sendOnelineMsg(210, 3, data) <= 0);
    if (where == PRESS)
      cancelDialogButton();
  }
  if (logfileon && where != NOLOG)
    fprintf(logfp, "%s", data);
}

#else
/*
 *  Print a line. 'Where' controls where msg goes.
 *  NOPRESS = No 'Press key'; PRESS = 'Press key'; LOG = logfile only; NOLOG
 */
void iomgr::printline(int where, mstring data) {
  if (where != LOG) {
    printf("%s", data);
    if (where == PRESS) 
      printf("%c\n", 7);
  }
  if (logfileon && where != NOLOG)
    fprintf(logfp, "%s", data);
}

void iomgr::openDialog(int nrlines){
  nrlines = 0;
}

void iomgr::closeDialog() {
}

void iomgr::cancelDialogButton() {
}
#endif

void iomgr::drawnode(ipaddr addr, nametype name, int snmp, int ops, int x,
		     int y, int typ) {
  int m1 = 201, m2;
  msgstr msg;
  switch (typ) {
    case ROUTER: m2 = 2; break;
    case HOST:   m2 = 2; break;
    case UNK:;
    case GATEWAY:;
    default: m2 = 1; break;
  }
  sprintf(msg, "%s %s %d %d %d %d", addr, name, snmp, ops, x, y);
  while(msgMgr.sendOnelineMsg(m1, m2, msg) <= 0);
}

void iomgr::drawlink(ipaddr a, ipaddr b) {
  int m1 = 202, m2 = 1;
  msgstr msg;
  sprintf(msg, "%s %s", a, b);
  while(msgMgr.sendOnelineMsg(m1, m2, msg) <= 0);
}

void iomgr::drawLAN(ipaddr addr, int intfc, int x, int y) {
  int m1 = 201, m2 = 4;
  msgstr msg;
  nametype name;
  switch (intfc) {
  case 2: strcpy(name, "regular1822"); break;
  case 3: strcpy(name, "hdh1822"); break;
  case 4: strcpy(name, "ddn-x25"); break;
  case 5: strcpy(name, "rfc877-x25"); break;
  case 6: strcpy(name, "ethernet-csmacd"); break;
  case 7: strcpy(name, "88023-csmacd"); break;
  case 8: strcpy(name, "88024-tokenbus"); break;
  case 9: strcpy(name, "88025-tokenring"); break;
  case 10: strcpy(name, "88026-man"); break;
  case 11: strcpy(name, "starLAN"); break;
  case 12: strcpy(name, "proteon-10Mb"); break;
  case 13: strcpy(name, "proteon-80Mb"); break;
  case 14: strcpy(name, "hyperchannel"); break;
  case 15: strcpy(name, "fddi"); break;
  case 16: strcpy(name, "lapb"); break;
  case 17: strcpy(name, "sdlc"); break;
  case 18: strcpy(name, "t1-carrier"); break;
  case 19: strcpy(name, "cept"); break;
  case 20: strcpy(name, "basicISDN"); break;
  case 21: strcpy(name, "primaryISDN"); break;
  case 22: strcpy(name, "Pt-PtSerial"); break;
  case 1:;
  default: strcpy(name, "other"); break;
  }
  sprintf(msg, "%s %s %d %d %d %d", addr, name, YES, YES, x, y);
  while (msgMgr.sendOnelineMsg(m1, m2, msg) <= 0);
}

void iomgr::undrawnode(ipaddr addr) {
  int m1 = 203, m2 = 1;
  while(msgMgr.sendOnelineMsg(m1, m2, addr) <= 0);
}

void iomgr::undrawlink(ipaddr a, ipaddr b) {
  int m1 = 204, m2 = 1;
  msgstr msg;
  sprintf(msg, "%s %s", a, b);
  while(msgMgr.sendOnelineMsg(m1, m2, msg) <= 0);
}

void iomgr::undrawLAN(ipaddr addr) {
  int m1 = 203, m2 = 4;
  while (msgMgr.sendOnelineMsg(m1, m2, addr) <= 0);
}

void iomgr::drawNormalBorder(ipaddr addr, int typ) {
  int m1 = 206, m2 = 2;
  msgstr m;
  sprintf(m, "%s %d", addr, typ);
  while(msgMgr.sendOnelineMsg(m1, m2, m) <= 0);
}

void iomgr::drawMonitorBorder(ipaddr addr, int typ) {
  int m1 = 206, m2 = 1;
  msgstr m;
  sprintf(m, "%s %d", addr, typ);
  while(msgMgr.sendOnelineMsg(m1, m2, m) <= 0);
}

void iomgr::newStatus(ipaddr addr, int snmp, int ops) {
  int m1 = 205, m2 = 1;
  msgstr msg;
  sprintf(msg, "%s %d %d", addr, snmp, ops);
  while (msgMgr.sendOnelineMsg(m1, m2, msg) <= 0);
}
