#include "lclworld.h"
#include <string.h>

extern int findnetname(char*, char*);
extern int findnetaddr(char*, char*);

 lclworld::lclworld() {
   nrRec = 0;
   lclworldptr[0] = NULL;
 }

/*
 *  Open, read and close the file "LclWorld" containing the name/addr/community
 *    info.  Put data in lclworldptr[] & return no. records read.
 *    Return -1 if an error occurs opening or reading file.
 */
int lclworld::loadworld() {
  FILE *fp;
  char inputstr[512], *aliasptr;
  nametype alias, s, c;
  addrlist *newaliasptr;
  int nrRead = 0;
  if ((fp = fopen("LclWorld.data","r")) == NULL) {
    return -1;
  }
  while (!feof(fp) && nrRec <= MAXLCLWORLDVARS) {
    if ((fgets(inputstr, 512, fp)) != NULL && !feof(fp)) {
      lclworldptr[nrRec] = new lclworldvar;
      lclworldptr[nrRec]->aliasptr = NULL;
      nrRead = sscanf(inputstr,"%s %s %s %s %s",lclworldptr[nrRec]->addr, 
		      lclworldptr[nrRec]->name, lclworldptr[nrRec]->cid, c, s);
      if (strncmp(s, "SNMP", 4) == 0) 
	lclworldptr[nrRec]->snmp = YES;
      else
	if (strncmp(s, "NOSNMP", 6) == 0)
	  lclworldptr[nrRec]->snmp = NO;
	else
	  lclworldptr[nrRec]->snmp = UNK;
      if (strncmp(c, "GATE", 4) == 0) 
	lclworldptr[nrRec]->ctype = GATEWAY;
      else
	if (strncmp(c, "ROUT", 4) == 0)
	  lclworldptr[nrRec]->ctype = ROUTER;
	else
	  if (strncmp(c, "HOST", 4) == 0)
	    lclworldptr[nrRec]->ctype = HOST;
	  else
	    lclworldptr[nrRec]->ctype = UNK;
      if (nrRead == 5) {
	aliasptr = strchr(inputstr, ':');
	if (aliasptr != NULL) {
	  aliasptr += 1;
	  while (nrRead > 0) {
	    nrRead = sscanf(aliasptr, " %s", alias);
	    if (nrRead > 0) {
	      newaliasptr = new addrlist;
	      strcpy(newaliasptr->addr, alias);
	      newaliasptr->next = lclworldptr[nrRec]->aliasptr;
	      lclworldptr[nrRec]->aliasptr = newaliasptr;
	      aliasptr = aliasptr + strlen(alias) + 1;
	    }
	  }
	}
	nrRec++;
      }
      else
	delete lclworldptr[nrRec];
    }
  }
  fclose(fp);
  return nrRec;
}

int lclworld::saveworld() {
  FILE *fp;
  int recno;
  addrlist *ptr;
  if ((fp = fopen("LclWorld.data","w")) == NULL)
    return -1;
  for (recno = 0; recno < nrRec; recno++) {
    fprintf(fp, "%s %s %s ", lclworldptr[recno]->addr, 
	    lclworldptr[recno]->name, lclworldptr[recno]->cid);
    switch(lclworldptr[recno]->ctype) {
      case ROUTER: fprintf(fp,"ROUTER "); 
           break;
      case GATEWAY: fprintf(fp,"GATEWAY ");
           break;
      case HOST: fprintf(fp,"HOST ");
           break;
      case UNK:;
      default: fprintf(fp,"UNK ");
           break;
      }
    switch(lclworldptr[recno]->snmp) {
      case YES: fprintf(fp,"SNMP");
           break;
      case NO: fprintf(fp,"NOSNMP");
           break;
      case UNK:; 
      default: fprintf(fp,"UNKSNMP");
           break;
      }
    if (lclworldptr[recno]->aliasptr != NULL) {
      fprintf(fp, ":");
      ptr = lclworldptr[recno]->aliasptr;
      while (ptr != NULL) {
	fprintf(fp," %s", ptr->addr);
	ptr = ptr->next;
      }
    }
    fprintf(fp,"\n");
    delete lclworldptr[recno];
  }
  fclose(fp);
  return nrRec;
}

void lclworld::printworld() {
  int i;
  addrlist *ptr;
  for (i = 0; i < nrRec; i++) {
    printf("ipaddr:   %s\n", lclworldptr[i]->addr);
    printf("  name:   %s\n", lclworldptr[i]->name);
    printf("  commun: %s\n", lclworldptr[i]->cid);
    printf("  snmp:   %d\n", lclworldptr[i]->snmp);
    printf("  ctype:  %d\n", lclworldptr[i]->ctype);
    ptr = lclworldptr[i]->aliasptr;
    while (ptr != NULL) {
      printf("  alias:  %s\n", ptr->addr);
      ptr = ptr->next;
    }
  }
}

/*
 *  Insert rec into the LclWorld database.
 */
int lclworld::insertrec(ipaddr a) {
  int recno;
  if (nrRec == MAXLCLWORLDVARS) return -1;
  recno = nrRec;
  while (recno > 0 && (strcmp(a, lclworldptr[recno-1]->addr) < 0)) {
    lclworldptr[recno] = lclworldptr[recno-1];
    recno--;
  }
  lclworldptr[recno] = new lclworldvar;
  strcpy(lclworldptr[recno]->addr, a);
  nrRec++;
  return recno;
}

/*
 *  Find the addr in the LclWorld array.
 *  Return indexnr+1 if found; 0 if not found; -1 if file error;
 */
int lclworld::findaddr(ipaddr addr) {
  int i, low = 0, mid, hi;
  hi = nrRec - 1;
  while (low <= hi) {
    mid = (low + hi)/2;
    i = strcmp(addr, lclworldptr[mid]->addr);
    if (i < 0) hi = mid - 1;
    else 
      if (i > 0) low = mid + 1;
      else return mid + 1;
  }
  return 0;
}

/*
 *  Find the addr in the LclWorld array for which alias is given.
 *  Return indexnr+1 if found; 0 if not found; -1 if file error;
 */
int lclworld::findalias(ipaddr addr) {
  int i;
  addrlist *ptr;
  for (i = 0; i < nrRec; i++) {
    ptr = lclworldptr[i]->aliasptr;
    while (ptr != NULL) {
      if (strcmp(ptr->addr, addr) == 0)
	return i+1;
      ptr = ptr->next;
    }
  }
  return 0;
}

/*
 *  Add new address to LclWorld if not already in file.  If on file return >0
 *    and actual address.  If not, get name, then addr again an check for
 *    aliases.  Add new alias or new name as appropriate and return 0.
 */
int lclworld::addtoworld(ipaddr addr) {
  ipaddr a;
  nametype name;
  int i;
  if ((i = findaddr(addr)) > 0)
    return i;
  if ((i = findalias(addr)) > 0) {
    strcpy(addr, lclworldptr[i-1]->addr);  /* return real addr, not alias */
    return i;
  }
  i = findnetname(name, addr);
  if (i < 0 || (strlen(name) == 0)) 
    return 0;                   /* only an address so don't make table entry */
  i = findnetaddr(name, a);
  if ((i < 0) || (strlen(a) == 0) || (strcmp(a, addr) == 0)){
    i = insertrec(addr);      /* insert with address provided */
    lclworldptr[i]->aliasptr = NULL;
  }
  else { 
    i = insertrec(a);   /* addr was alias; use new one from nameserver */
    lclworldptr[i]->aliasptr = new addrlist;
    strcpy(lclworldptr[i]->aliasptr->addr, addr);
    lclworldptr[i]->aliasptr->next = NULL;
  }
  strcpy(lclworldptr[i]->name, name);
  strcpy(lclworldptr[i]->cid, "public");
  lclworldptr[i]->snmp = UNK;
  lclworldptr[i]->ctype = UNK;
  return nrRec;
}
/*
 *  Given the ip addr, set the name.  Return 0 if addr not found, 1 if done.
 */
int lclworld::setworldName(nametype name, ipaddr addr) {
  int i;
  if ((i = findaddr(addr)) <= 0)
    return 0;
  strcpy(lclworldptr[i-1]->name, name);
  return i;
}

/*
 *  Given the ip addr, set community.  Return 0 if addr not found, 1 if done.
 */
int lclworld::setworldCid(commid cid, ipaddr addr) {
  int i;
  if ((i = findaddr(addr)) <= 0)
    return 0;
  strcpy(lclworldptr[i-1]->cid, cid);
  return i;
}

/*
 *  Given the ip addr, set snmp status.  Return 0 if addr not found, 1 if done.
 */
int lclworld::setworldSnmp(int snmp, ipaddr addr) {
  int i;
  if ((i = findaddr(addr)) <= 0)
    return 0;
  lclworldptr[i-1]->snmp = snmp;
  return i;
}

/*
 *  Given the ip addr, set the type.  Return 0 if addr not found, 1 if done.
 */
int lclworld::setworldType(int ctype, ipaddr addr) {
  int i;
  if ((i = findaddr(addr)) <= 0)
    return 0;
  lclworldptr[i-1]->ctype = ctype;
  return i;
}

/*
 *  Given the ip addr add an alias.  Return 0 if addr not found, 1 if done.
 */
int lclworld::setworldAlias(ipaddr alias, ipaddr addr) {
  int i;
  addrlist *ptr;
  if ((i = findaddr(addr)) <= 0)
    return 0;
  ptr = lclworldptr[i-1]->aliasptr;
  while (ptr != NULL) {
    if (strcmp(ptr->addr, alias) == 0)  /* already there */
      return i;
    ptr = ptr->next;
  }
  ptr = new addrlist;                   /* add in new alias */
  strcpy(ptr->addr, alias);
  ptr->next = lclworldptr[i-1]->aliasptr;
  lclworldptr[i-1]->aliasptr = ptr;
  return i;
}

/*
 *  Given the addr, return corresponding name. rtn 0 if addr not in tbl, else 1
 */
int lclworld::getworldNamefmAddr(nametype n, ipaddr a) {
  int i;
  n[0] = '\0';
  if ((i = findaddr(a)) <= 0)
    return 0;
  strcpy(n, lclworldptr[i-1]->name);
  return 1;
}

/*
 *  Given the addr, return corresponding comm. rtn 0 if addr not in tbl, else 1
 */
int lclworld::getworldCidfmAddr(commid cid, ipaddr a) {
  int i;
  cid[0] = '\0';
  if ((i = findaddr(a)) > 0) {
    strcpy(cid, lclworldptr[i-1]->cid);
    return 1;
  }
  if ((i = findalias(a)) > 0) {
    strcpy(cid, lclworldptr[i-1]->cid);
    return 1;
  }
  return 0;
}

/*
 *  Given the addr, return snmp value. rtn 0 if addr not in tbl, else 1
 */
int lclworld::getworldSnmpfmAddr(int& s, ipaddr a) {
  int i;
  s = 0;
  if ((i = findaddr(a)) <= 0)
    return 0;
  s = lclworldptr[i-1]->snmp;
  return 1;
}

/*
 *  Given the addr, return type value. rtn 0 if addr not in tbl, else 1
 */
int lclworld::getworldTypefmAddr(int& t, ipaddr a) {
  int i;
  t = 0;
  if ((i = findaddr(a)) <= 0)
    return 0;
  t = lclworldptr[i-1]->ctype;
  return 1;
}

/*
 *  Given an alias, return corresponding addr. rtn 0 if addr not in tbl, else 1
 */
int lclworld::getworldAddrfmAlias(ipaddr addr, ipaddr alias) {
  int i;
  addr[0] = '\0';
  if ((i = findalias(alias)) <= 0)
    return 0;
  strcpy(addr, lclworldptr[i-1]->addr);
  return 1;
}

/*
 *  Given a name, find the addr in lclworld.  rtn 0 if not found, else 1.
 */
int lclworld::getworldAddrfmName(nametype name, ipaddr addr) {
  int i;
  for (i = 0; i < nrRec; i++) {
    if (strcmp(lclworldptr[i]->name, name) == 0) {
      strcpy(addr, lclworldptr[i]->addr);
      return 1;
    }
  }
}

/* for testing of this object
main(){
  ipaddr addr, alias;
  int i;
  lclworld l;
  l.loadworld();
  strcpy(alias, "130.42.46.6");
  i = l.getworldAddrfmAlias(addr, alias);
  if (i) printf("addr: %s\n", addr );
  else printf("Trouble\n");
  l.printworld();
  l.saveworld();
}
*/
