#include <stdio.h>
#include "../msg/snmpmsg.h"
#include "actmgr.h"
#include <time.h>
#include <string.h>

/* #define DEBUG 0 */

extern int snmpmsg(int, ipaddr, commid, vartable&, long, responseList&, int);
extern long time(long);
extern char* ctime(long*);
extern int findnetname(nametype, ipaddr);
extern int findnetaddr(nametype, ipaddr);

struct interface {
  int nr;              /* the local interface nr */
  int typ;             /* MIB _ifType code */
  mstring descr;       /* ascii of interface type */
  int nrOnNet;         /* nr components on direct link to this interface */
  unsigned long speed;
  ipaddr subnet;       /* the subnet for the interface */
  addrlist *adjptr;    /* list of all components connected */
} *intfcptr[50];

int sendamsg(int typ, ipaddr addr, commid cid, vartable& mibvar,
	     responseList& resp, int repeats, int delay) {
  int errno = -1, i = 0;
  while (errno != 0 && i < repeats) {
    errno = snmpmsg(typ, addr, cid, mibvar, delay, resp, RETURNONERROR);
    if (errno < 0)
      delay += delay/2;
    i++;
  }
  return errno;
}

 actmgr::actmgr() {
   memNum = 0;
 }

void actmgr::getsubnet(ipaddr addr, ipaddr subnet, int mask) {
  int i = 0, cntr = 0;
  strcpy(subnet, "0.0.0.0");
  if (mask < 0 || mask > 4) return;
  if (mask == 4) {
    strcpy(subnet, addr);
    return;
  }
  while (i < mask && cntr < 18) {
    if (addr[cntr] == '.') i++;
    cntr++;
  }
  if (cntr > 17) return;
  strncpy(subnet, addr, cntr);
  subnet[cntr]= '\0';
  if (mask == 1) strcat(subnet, "0.0.0");
  else
    if (mask == 2) strcat(subnet, "0.0");
    else strcat(subnet, "0");
}

void actmgr::opstat(mstring inputdata, listmgr& lmgr, iomgr& io) {
  mstring data, str;
  nametype name;
  ipaddr addr, alias;
  commid cid;
  int snmp, typ, addrinworld, wasOperational;
  gateway *g;
  io.openDialog(5);
  sscanf(inputdata, "%s%s", addr, name);
  if ((addrinworld = lmgr.findNameAddr(addr,alias,name,cid,snmp,typ)) == 0) {
    sprintf(str, "Cannot monitor %s (%s)", alias, name);
    io.printline(PRESS, str);
    return;
  }
  if ((g = lmgr.gtwy_find(addr)) != NULLPTR) {
    wasOperational = g->getopstat();
    if (g->operational()) {
      sprintf(data, "%s\t(%s) is operational\n", addr, g->getname());
      if (wasOperational != 1) io.newStatus(addr, YES, YES);
    }
    else {
      sprintf(data, "%s\t(%s) not responding\n", addr, g->getname());
      if (wasOperational == 1) io.newStatus(addr, snmp, NO);
    }
  }
  else
    sprintf(data, "%s not in network\n", addr);
  io.printline(PRESS, data);
}

void actmgr::monitor(mstring str, listmgr& lmgr, iomgr& io) {
  ipaddr addr, alias;
  nametype name;
  commid cid;
  int interval = 0, addrinworld, snmp, typ;
  gateway *g;
  sscanf(str, "%s%s%d", addr, name, &interval);
  if ((addrinworld = lmgr.findNameAddr(addr,alias,name,cid,snmp,typ)) == 0) {
    io.openDialog(5);
    sprintf(str, "Cannot monitor %s (%s)", alias, name);
    io.printline(PRESS, str);
    return;
  }
  if ((g = lmgr.gtwy_find(addr)) != NULLPTR) {
    if (interval < 0 || interval > 86400)
      interval = 0;             /* paranoia - no interval > 24 hours */
    g->set_monitor(interval);
    if (interval == 0) io.drawNormalBorder(addr, g->gettype());
    else io.drawMonitorBorder(addr, g->gettype());
  }
  else {
    io.openDialog(4);
    sprintf(str, "%s not in network\n", addr);
    io.printline(PRESS, str);
  }
}

void actmgr::queryone(mstring inputdata, listmgr& lmgr, iomgr& io) {
  ipaddr addr;
  mstring data;
  commid comm, cid;
  int errno;
  vartable mibvar;
  responseList response;
  mibvar.nr = 1;
  response.vars.nr = 0;
  sscanf(inputdata, "%s%s%s", addr, cid, data);
  strcpy(mibvar.name[0], data);
  lmgr.findcommunity(comm, addr);
  if (strcmp(comm, "public") == 0 && strcmp(cid, "public") != 0)
    strcpy(comm, cid);
  io.openDialog(9);
  sprintf(inputdata, "querying %s...\n\n", addr);
  io.printline(NOPRESS, inputdata);
  errno = sendamsg(REQ, addr, comm, mibvar, response, 2, 6);
  if (errno == 0) {
    sprintf(inputdata, "%s: %s\n", response.vars.name[0], 
	    response.asciival[0]);
    io.printline(NOPRESS, inputdata);
  }
  else {
    sprintf(inputdata, "No response for %s: %s\n", addr, data);
    io.printline(NOPRESS, inputdata);
  }
  io.printline(PRESS,"\n");
}

void actmgr::querymany(mstring inputdata, listmgr& lmgr, iomgr& io) {
  ipaddr addr;
  mstring data;
  commid comm, cid;
  int errno, answered = 0, toofar;
  vartable mibvar;
  responseList response;
  mibvar.nr = 1;
  response.vars.nr = 0;
  sscanf(inputdata, "%s%s%s", addr, cid, data);
  strcpy(mibvar.name[0], data);
  lmgr.findcommunity(comm, addr);
  if (strcmp(comm, "public") == 0 && strcmp(cid, "public") != 0)
    strcpy(comm, cid);
  io.openDialog(24);
  sprintf(inputdata, "querying %s...\n\n", addr);
  io.printline(NOPRESS, inputdata);
  do {
    errno = sendamsg(NXT, addr, comm, mibvar, response, 3, 5);
    if (errno == 0) {
      strcpy(mibvar.name[0], response.vars.name[0]);
      toofar = strncmp(mibvar.name[0], data, strlen(data));
      if (!toofar) {
	sprintf(inputdata, "%s: %s\n", response.vars.name[0], 
		response.asciival[0]);
	io.printline(NOPRESS, inputdata);
	answered = 1;
      }
    }
  }
  while (!errno && !toofar);
  if (errno) {
    if (!answered) {
      sprintf(inputdata, "No response for %s: %s\n", addr, data);
      io.printline(NOPRESS, inputdata);
    }
    else
      io.printline(NOPRESS, "Loss of response....incomplete data\n");
  }
  io.printline(PRESS,"\n");
}

void actmgr::printdetlist(listmgr& lmgr, iomgr& io) {
}

void actmgr::printcomponent(mstring addr, listmgr& lmgr, iomgr& io) {
  mstring data;
  ipaddr d;
  gateway *g;
  long currtime;
  int i;
  g = lmgr.gtwy_find(addr);
  if (g == NULL) {
    sprintf(data, "%s not in network.\n", addr);
    io.printline(NOPRESS, data);
  }
  else {
    currtime = time(NULL);
    sprintf(data, "Component Listing at: %-25.25s", ctime(&currtime));
    io.printline(NOPRESS, data);
    sprintf(data, "  Address:     %s\n", addr);
    io.printline(NOPRESS, data);
    sprintf(data, "  Name:        %s\n", g->getname());
    io.printline(NOPRESS, data);
    sprintf(data, "  Community:   %s\n", g->getcomm());
    io.printline(NOPRESS, data);
    if (g->getopstat())
      io.printline(NOPRESS, "  Operational: Yes\n");
    else
      io.printline(NOPRESS, "  Operational: No\n");
    i = g->gettype();
    switch(i) {
    case GATEWAY:
      sprintf(data, "  Type:        Gateway\n");
      break;
    case ROUTER:
      sprintf(data, "  Type:        Router\n");
      break;
    case HOST:
      sprintf(data, "  Type:        Host\n");
      break;
    case UNK:;
    default:
      sprintf(data, "  Type:        Unknown\n");
      break;
    }
    io.printline(NOPRESS, data);
    i = g->getsnmp();
    switch(i) {
    case YES:
      sprintf(data, "  SNMP:        Yes\n");
      break;
    case NO:
      sprintf(data, "  SNMP:        No\n");
      break;
    case UNK:;
    default:
      sprintf(data, "  SNMP:        Unknown\n");
      break;
    }
    io.printline(NOPRESS, data);
    sprintf(data, "  X,Y Coord:   %d, %d\n", g->getxcoord(), g->getycoord());
    io.printline(NOPRESS, data);
  }
  i = g->getfirstneighbor(d);
  while (i) {
    sprintf(data, "  neighbor:    %s\n",d);
    io.printline(NOPRESS, data);
    i = g->getnextneighbor(d);
  }
  io.printline(PRESS, "\n");
}

void actmgr::printnodes(listmgr& lmgr, iomgr& io) {
  mstring data;
  long currtime;
  gateway *g, *gfirst;
  char s, t, o;
  int i;
  gfirst = lmgr.gtwy_getfirst();
  if (gfirst == NULL) return;
  io.openDialog(24);
  g = gfirst;
  currtime = time(NULL);
  sprintf(data, "Node Summary at:  %-35.35s\n", ctime(&currtime));
  io.printline(NOPRESS, data);
  strcpy(data,"IP address       Name                                ");
  strcat(data, "SNMP  Type  Oper\n");
  io.printline(NOPRESS, data);
  strcpy(data, "----------------------------------------------------");
  strcat(data,"-----------------\n");
  io.printline(NOPRESS, data);
  do {
    s = t = o = ' ';
    if (g->getopstat())
      o = 'Y';
    i = g->getsnmp();
    if (i == UNK) s = 'U';
    else
      if (i == YES) s = 'Y';
      else s = 'N';
    i = g->gettype();
    switch(i) {
    case GATEWAY: t = 'G';
      break;
    case ROUTER:  t = 'R';
      break;
    case HOST:    t = 'H';
      break;
    case UNK:;
    default:      t = 'U';
      break;
    }
    sprintf(data, "%-16.16s %-35.35s   %c     %c     %c\n", g->getaddr(),
	    g->getname(), s, t, o);
    io.printline(NOPRESS, data);
    g = lmgr.gtwy_getnext();
  }
  while (g != gfirst);
  io.printline(PRESS, "\n");
}

void actmgr::printlinks(listmgr& lmgr, iomgr& io) {
  mstring data, proto, typ;
  ipaddr addr1, addr2;
  long currtime;
  link *l, *lfirst;
  int p, t;
  unsigned long speed;
  lfirst = lmgr.link_getfirst();
  if (lfirst == NULLPTR) return;
  io.openDialog(24);
  l = lfirst;
  currtime = time(NULL);
  sprintf(data, "Link Summary at:  %-35.35s\n", ctime(&currtime));
  io.printline(NOPRESS, data);
  strcpy(data,"From address     To Address       Link Type        ");
  strcat(data, "Protocol   Speed(bps)\n");
  io.printline(NOPRESS, data);
  strcpy(data, "----------------------------------------------------");
  strcat(data,"--------------------\n");
  io.printline(NOPRESS, data);
  do {
    speed = l->getintfcSpeed();
    t = l->getintfcType();
    p = l->getrouteProto();
    switch(t) {
    case 2: strcpy(typ, "regular1822"); break;
    case 3: strcpy(typ, "hdh1822"); break;
    case 4: strcpy(typ, "ddn-x25"); break;
    case 5: strcpy(typ, "rfc877-x25"); break;
    case 6: strcpy(typ, "ethernet-csmacd"); break;
    case 7: strcpy(typ, "88023-csmacd"); break;
    case 8: strcpy(typ, "88024-tokenbus"); break;
    case 9: strcpy(typ, "88025-tokenring"); break;
    case 10: strcpy(typ, "88026-man"); break;
    case 11: strcpy(typ, "starLAN"); break;
    case 12: strcpy(typ, "proteon-10Mb"); break;
    case 13: strcpy(typ, "proteon-80Mb"); break;
    case 14: strcpy(typ, "hyperchannel"); break;
    case 15: strcpy(typ, "fddi"); break;
    case 16: strcpy(typ, "lapb"); break;
    case 17: strcpy(typ, "sdlc"); break;
    case 18: strcpy(typ, "t1-carrier"); break;
    case 19: strcpy(typ, "cept"); break;
    case 20: strcpy(typ, "basicISDN"); break;
    case 21: strcpy(typ, "primaryISDN"); break;
    case 22: strcpy(typ, "Pt-Pt Serial"); break;
    case 1:;
    default: strcpy(typ, "other"); break;
    }
    switch (p) {
    case 2: strcpy(proto, "local"); break;
    case 3: strcpy(proto, "netmgmt"); break;
    case 4: strcpy(proto, "icmp"); break;
    case 5: strcpy(proto, "egp"); break;
    case 6: strcpy(proto, "ggp"); break;
    case 7: strcpy(proto, "hello"); break;
    case 8: strcpy(proto, "rip"); break;
    case 9: strcpy(proto, "is-is"); break;
    case 10: strcpy(proto, "es-is"); break;
    case 11: strcpy(proto, "ciscoIgrp"); break;
    case 12: strcpy(proto, "bbnSpfIgrp"); break;
    case 13: strcpy(proto, "oigp"); break;
    case 1:;
    default: strcpy(proto, "other"); break;
    }
    l->getaddr(addr1, addr2);
    sprintf(data, "%-16.16s %-16.16s %-15.15s  %-9.9s %11lu\n", addr1, addr2,
	    typ, proto, speed);
    io.printline(NOPRESS, data);
    l = lmgr.link_getnext();
  }
  while (l != lfirst);
  io.printline(PRESS, "\n");
}

void actmgr::printLANs(listmgr& lmgr, iomgr& io) {
  mstring data, typ, blanks;
  ipaddr name, addr;
  long currtime;
  LAN *lan, *lanfirst;
  int i, t, intfc, x, y;
  unsigned long speed;
  lanfirst = lan = lmgr.LAN_getfirst();
  if (lanfirst == NULLPTR) return;
  io.openDialog(24);
  currtime = time(NULL);
  sprintf(data, "LAN Summary at:  %-35.35s\n", ctime(&currtime));
  io.printline(NOPRESS, data);
  strcpy(data,"LAN Name         Link Type         Speed(bps)");
  strcat(data, "   Member Address\n");
  io.printline(NOPRESS, data);
  strcpy(data, "--------------------------------------------");
  strcat(data,"-------------------\n");
  io.printline(NOPRESS, data);
  strcpy(blanks,"                                               ");
  do {
    speed = lan->getintfcSpeed();
    t = lan->getintfcType();
    switch(t) {
    case 2: strcpy(typ, "regular1822"); break;
    case 3: strcpy(typ, "hdh1822"); break;
    case 4: strcpy(typ, "ddn-x25"); break;
    case 5: strcpy(typ, "rfc877-x25"); break;
    case 6: strcpy(typ, "ethernet-csmacd"); break;
    case 7: strcpy(typ, "88023-csmacd"); break;
    case 8: strcpy(typ, "88024-tokenbus"); break;
    case 9: strcpy(typ, "88025-tokenring"); break;
    case 10: strcpy(typ, "88026-man"); break;
    case 11: strcpy(typ, "starLAN"); break;
    case 12: strcpy(typ, "proteon-10Mb"); break;
    case 13: strcpy(typ, "proteon-80Mb"); break;
    case 14: strcpy(typ, "hyperchannel"); break;
    case 15: strcpy(typ, "fddi"); break;
    case 16: strcpy(typ, "lapb"); break;
    case 17: strcpy(typ, "sdlc"); break;
    case 18: strcpy(typ, "t1-carrier"); break;
    case 19: strcpy(typ, "cept"); break;
    case 20: strcpy(typ, "basicISDN"); break;
    case 21: strcpy(typ, "primaryISDN"); break;
    case 22: strcpy(typ, "Pt-Pt Serial"); break;
    case 1:;
    default: strcpy(typ, "other"); break;
    }
    strcpy(name, lan->getname());
    i = lan->getfirstmember(addr, intfc, x, y);
    sprintf(data, "%-16.16s %-15.15s  %11lu   %-16.16s\n", name, typ, speed,
	    addr);
    io.printline(NOPRESS, data);
    i = lan->getnextmember(addr, intfc, x, y);
    while (i != 0) {
      sprintf(data, "%s %s\n", blanks, addr);
      io.printline(NOPRESS, data);
      i = lan->getnextmember(addr, intfc, x, y);
    }
    lan = lmgr.LAN_getnext();
  }
  while (lan != lanfirst);
  io.printline(PRESS, "\n");
}

int actmgr::notonneighborlist(addrlist* ptr, ipaddr addr) {
  while (ptr != NULLPTR) {
    if (strcmp(ptr->addr, addr) == 0) return 0;
    ptr = ptr->next;
  }
  return 1;
}

/*
 *  Send messages to Xio to draw the network after loading.
 */
void actmgr::drawnetwork(listmgr& lmgr, iomgr& io) {
  gateway *g, *g1;
  LAN *lan, *lan1;
  link *l, *l1;
  int x, y, intfc, i;
  ipaddr addr, lanaddr, a, b;
  g = g1 = lmgr.gtwy_getfirst();   /* draw nodes first */
  if (g == NULLPTR) return;        /*  unless network empty */
  do {
    io.drawnode(g->getaddr(), g->getname(), g->getsnmp(), g->getopstat(),
		g->getxcoord(), g->getycoord(), g->gettype());
    g = lmgr.gtwy_getnext();
  }
  while(g != g1);
  lan = lan1 = lmgr.LAN_getfirst();
  if (lan != NULLPTR) {
    do {
      lan->getLANcoord(x,y);
      io.drawLAN(lan->getname(), lan->getintfcType(), x, y);
      lan = lmgr.LAN_getnext();
    }
    while (lan != lan1);
    lan = lmgr.LAN_getfirst();
    strcpy(lanaddr, lan->getname());
    do {
      i = lan->getfirstmember(addr, intfc, x,y);
      while (i != 0) {
	if (lmgr.gtwy_find(addr) != NULLPTR)
	  io.drawlink(lanaddr, addr);
	i = lan->getnextmember(addr, intfc, x,y);
      }
      lan = lmgr.LAN_getnext();
      strcpy(lanaddr, lan->getname());      
    }
    while (lan != lan1);
  }
  l = l1 = lmgr.link_getfirst();
  if (l == NULLPTR) return;
  do {
    l->getaddr(a,b);
    if (lmgr.gtwy_find(a) != NULLPTR && lmgr.gtwy_find(b) != NULLPTR)
      io.drawlink(a,b);
    l = lmgr.link_getnext();
  }
  while (l != l1);
}

/*
 *  Update node coordinates after icon move.
 */
void actmgr::moveNode(msgstr inputstr, listmgr& lmgr) {
  LAN *lan;
  gateway *g;
  link *l;
  ipaddr addr, a, b;
  int x, y, i, j, x1, y1;
  sscanf(inputstr, "%s%d%d", addr, &x, &y);
  if ((g = lmgr.gtwy_find(addr)) == NULLPTR) {
/*%%%%printf("  Can't find %s to move\n", addr); */
    return;
  }
  g->setcoords(x,y);
/*%%%%%printf("  Moving %s to %d,%d\n", addr, x,y); */
  i = g->getfirstneighbor(a);
  while (i != 0) {                /* look through the node's neighbor list */
    if (strncmp(a, "LAN", 3) == 0) {
      if ((lan = lmgr.LAN_find(a)) != NULLPTR)  /* move coords on LAN list */

	j = lan->mv_member(addr, x, y);
    }
    else {                             /* must be a link connnection */
      if (strcmp(addr, a) < 0) {       /* links are (addr1 < addr2 order */
	if ((l = lmgr.link_find(addr, a)) != NULLPTR) {
	  l->getb_coord(x1, y1);       /* get coords for neighbor */     
	  l->setcoords(x, y, x1, y1);  /* and reset both */
	}
      }
      else {
	if ((l = lmgr.link_find(a, addr)) != NULLPTR) {
	  l->geta_coord(x1, y1);       /* get coords for neighbor */
	  l->setcoords(x1, y1, x, y);  /*  and reset both */
	}
      }
    }
    i = g->getnextneighbor(a);         /* look for more neighbors */
  }
}

/*
 *  Update coords for LAN icon after X move.
 */
void actmgr::moveLAN(msgstr inputstr, listmgr& lmgr) {
  LAN *lan;
  ipaddr addr;
  int x,y;
  sscanf(inputstr, "%s%d%d", addr, &x, &y);
  if ((lan = lmgr.LAN_find(addr)) == NULLPTR) return;
  lan->setLANcoord(x,y);
}

/*
 *  Add new component to network and resolve connections. Return
 *    0 = Added, -1 = nonsnmp, -2 = no addr found, -3 = already in network,
 *   -4 = No connections to network, -5 = msg response failure
 */
int actmgr::addcomponent(int errormode, mstring str, 
			 listmgr& lmgr, iomgr& io) {
  ipaddr addr, alias, netaddr, netalias, n, subnet, nalias;
  nametype name, newname, netname;
  commid community, cid, ncid;
  mstring st1, st2, st3, st4, st5;
  gateway *g, *gfirst;
  LAN *lan;
  responseList response;
  int snmp, comptype, xcoord, ycoord, addrfound = 0, i, j, j1, k, k1, added;
  int errno, nrOfIntfc, nx, ny, nrOfConnections = 0, netsnmp, nettyp;
  int metric, rtype, intfc, rproto, output1, output2, subnetmask;
  int notconnected, netintfc;
  addrlist *neighborlistptr, *addrptr, *neighborptr;
  vartable mibvar, mibvar1;
  if(errormode == QUIETMODE)            /* Print only to log file when */
    output1 = output2 = LOG;            /* procedure called with QUIET set */
  else {
    output1 = NOPRESS;
    output2 = PRESS;
    io.openDialog(24);
    memNum = 0;                         /* Used to set posits for scrn */
  }
#ifdef DEBUG
  output1 = NOPRESS;                    /* print everything in DEBUG mode */
  output2 = PRESS;
#endif
  sscanf(str, "%s%s%s", addr, name, community);
  mibvar.nr = 1;
  strcpy(mibvar.name[0], "system_sysDescr_0");   /* use to test for response */
  addrfound = lmgr.findNameAddr(addr, alias, name, cid, snmp, comptype);
  if (strcmp(addr, "none") == 0 && !addrfound) { 
    sprintf(str, "Cannot add - no addr for %s\n", name);
    io.printline(output2, str);
    return -2;
  }
  if ((g = lmgr.gtwy_find(addr)) != NULLPTR) {
    sprintf(str, "%s not added - already in present network\n", addr);
    io.printline(output2, str);
    return -3;
  }
  if (addrfound == 1 || addrfound == 2) strcpy(community, cid);
  if (strcmp(alias, addr) != 0) {
    sprintf(str, "Using %s in lieu of alias %s\n\n", addr, alias);
    io.printline(output1, str);
  }
  if (snmp == YES) i = 3;
  else i = 2; 
  sprintf(str, "Testing %s for response...\n", addr);
  io.printline(output1, str);
  errno = sendamsg(REQ, addr, community, mibvar, response, i, 4);
  if (errno == 0) {
    snmp = YES;
    io.printline(output1, "Response received...\n\n");
  }
  else {
    io.printline(output2, "Cannot add to network - no snmp response!\n");
    return -1;
  }

/* Found address, name, community.  Next, develop network adjacencies. */
  strcpy(mibvar.name[0],"interfaces_ifNumber_0");
  errno = sendamsg(REQ, addr, community, mibvar, response, 3, 6);
  if (errno) {
    io.printline(output2, "Cannot add to network - poor message response\n");
    return -5;
  }
  else {
    nrOfIntfc = response.val[0].value.intgr;
    sprintf(str, "%s has %d interfaces:\n", addr, nrOfIntfc);
    io.printline(output1, str);
  }
  if (nrOfIntfc > 49) nrOfIntfc = 49;
  intfcptr[0] = new interface;
  intfcptr[0]->nr = 0;             /* 0 used for default interface nr later */
  for (i = 1; i <= nrOfIntfc; i++) {
    mibvar.nr = 4;
    sprintf(mibvar.name[0], "interfaces_ifTable_ifEntry_ifIndex_%d", i);
    sprintf(mibvar.name[1], "interfaces_ifTable_ifEntry_ifDescr_%d", i);
    sprintf(mibvar.name[2], "interfaces_ifTable_ifEntry_ifType_%d", i);
    sprintf(mibvar.name[3], "interfaces_ifTable_ifEntry_ifSpeed_%d", i);
    j = sendamsg(REQ, addr, community, mibvar, response, 3, 6);
    intfcptr[i] = new interface;
    if (j == 0) {
      intfcptr[i]->nr = response.val[0].value.intgr;
      strcpy(intfcptr[i]->descr, response.asciival[1]);
      intfcptr[i]->typ = response.val[2].value.intgr;
      intfcptr[i]->speed = response.val[3].value.cntr;
    }
    else {
      intfcptr[i]->nr = i;
      strcpy(intfcptr[i]->descr, "unknown");
      intfcptr[i]->typ = 1;
      intfcptr[i]->speed = 0;
    }
    sprintf(str, "Interface:   %d\n", intfcptr[i]->nr);
    io.printline(output1, str);
    sprintf(str, "  Descr:     %s\n", intfcptr[i]->descr);
    io.printline(output1, str);
    sprintf(str, "  Type code: %d\n", intfcptr[i]->typ);
    io.printline(output1, str);
    sprintf(str, "  Speed:     %lu\n", intfcptr[i]->speed);
    io.printline(output1, str);
    intfcptr[i]->subnet[0] = '\0';
    mibvar.nr = 1;                   /* now look for adjacent components */
    response.vars.nr = 0;            /*   on this interface */
    sprintf(st1, "at_atTable_atEntry_atNetAddress_%d",i);
    strcpy(mibvar.name[0], st1);
    intfcptr[i]->nrOnNet = 0;          /* total nr on interface */
    intfcptr[i]->adjptr = NULLPTR;     /* the list of those adjacent */
    j = 1;                             /* =0 signals end of interface list */
    do {
      errno = sendamsg(NXT, addr, community, mibvar, response, 3, 6);
      if (errno == 0) {
	strcpy(mibvar.name[0], response.vars.name[0]);  /* setup for next */
	if (strncmp(mibvar.name[0], st1, strlen(st1)) == 0) {
	  intfcptr[i]->nrOnNet++;       /* another comp. on same interface */
	  strcpy(netaddr, response.asciival[0]);  /* get addr returned */
	  strcpy(netname, "none");
	  addrfound = lmgr.findNameAddr(netaddr, netalias, netname, cid,
					netsnmp, nettyp);
	  if ((g = lmgr.gtwy_find(netaddr)) != NULLPTR) { /* found adj. one */
	    nrOfConnections++;                      /* keep count of conn. */
	    addrptr = new addrlist;
	    strcpy(addrptr->addr, netaddr);         /* Now add this address */
	    addrptr->next = intfcptr[i]->adjptr;    /*  to the list for this */
	    intfcptr[i]->adjptr = addrptr;          /*  interface */
	    sprintf(str, "  %s is adjacent.\n", netaddr);
	    io.printline(output1, str);
	  }
	  if (strlen(intfcptr[i]->subnet) == 0) {
	    sprintf(intfcptr[i]->subnet, "LAN.%s",netalias);/* the subnet */
	    k = strlen(intfcptr[i]->subnet);             /* identifier for */
	    while (intfcptr[i]->subnet[k] != '.' && k > 0)/* the subnet */
	      intfcptr[i]->subnet[k--] = '\0';           /* attached to */
	    intfcptr[i]->subnet[k] = '\0';               /* this interface */
	  }
	}
	else j = 0;               /* no longer same interface */
      }
    }  /* end do - look for another neighbor on this interface? */
    while (errno == 0 && j == 1);
    if (strlen(intfcptr[i]->subnet) > 0) {
      sprintf(str, "  Subnet:    %s\n", intfcptr[i]->subnet);
      io.printline(output1, str);
    }
    sprintf(str, "  %d connected to this interface\n", intfcptr[i]->nrOnNet);
    io.printline(output1, str);
  }                               /* now move to next interface */

  /* Now add this new component to an LAN and neighbors' lists */
  mibvar.nr = 1;
  neighborlistptr = NULLPTR;      /* no neighbors yet for this new comp. */
  xcoord = 0;
  ycoord = 0;
  for (i = 1; i <= nrOfIntfc; i++) {
#ifdef DEBUG
    sprintf(str, "Check interface %d on subnet: %s\n", i, intfcptr[i]->subnet);
    io.printline(output1, str);
#endif
    if (intfcptr[i]->adjptr == NULLPTR) continue; /* no neighbors here */
    addrptr = intfcptr[i]->adjptr;               /* otherwise get 1st */
    do {                                   /* now loop - check each neighbor */
      strcpy(netaddr, addrptr->addr);      /* addr of neighbor */
#ifdef DEBUG
    sprintf(str, "Check neighbor: %s\n", netaddr);
    io.printline(output1, str);
#endif
      if ((g = lmgr.gtwy_find(netaddr)) == NULLPTR) {  /* access gtwy */
	addrptr = addrptr->next;                       /* or leave if !found */
	continue;
      }
      k = g->getfirstneighbor(n);          /* neighbor's 1st neighbor */
#ifdef DEBUG
    sprintf(str, "Check neighbor's neighbor: %s\n", n);
    io.printline(output1, str);
#endif
      added = 0;                           /* to see if LAN exists (=1 does) */
      while (k > 0 && added == 0) {
	if (strncmp(n, "LAN", 3) == 0) {   /* see if it is an LAN */
	  if ((lan = lmgr.LAN_find(n)) != NULLPTR) {
	    if (strcmp(n, intfcptr[i]->subnet) == 0) {
	      added = 1;
	      if (notonneighborlist(neighborlistptr, n)) { /* already added? */
		sprintf(str, "Attach to %s via interface %d\n", n, i);
		io.printline(output1, str);
		if (xcoord == 0) {
		  lan->getLANcoord(nx, ny);
		  if (QUIETMODE) {
		    xcoord = nx + 80 + memNum * 4;
		    ycoord = ny + (memNum++) * 2;
		  }
		  else {
		    xcoord = nx + 80;
		    ycoord = ny + 30;
		  }
		}                                   /* next, add to LAN list */
		lan->add_member(addr, intfcptr[i]->nr, xcoord, ycoord);
		neighborptr = new addrlist;
		strcpy(neighborptr->addr, n);    /* add LAN name to new list */
		neighborptr->next = neighborlistptr;
		neighborlistptr = neighborptr;   /* and put at head of list */
	      }
	    }
#ifdef DEBUG
	    else {
	      sprintf(str,"Do not attach to %s on comp. interface %d\n", n, i);
	      io.printline(output1, str);
	    }
#endif
	  }
	}
	k = g->getnextneighbor(n);
      }
      if (!added) {                      /* must be a new LAN */
	nx = g->getxcoord() + 80;        /* temporary spot for new LAN */
	if (QUIETMODE)
	  ny = g->getycoord() + (nrOfIntfc - intfcptr[i]->nr/2) * 40;
	else
	  ny = g->getycoord() + 30; 
	strcpy(newname, intfcptr[i]->subnet);
	k = lmgr.LAN_add(newname, intfcptr[i]->typ, intfcptr[i]->speed,
			 nx, ny);         /* create the new LAN */
        if (k != 0) {
	  io.printline(output1, "Unable to add new LAN - list full\n");
	  addrptr = addrptr->next;
	  nrOfConnections--;
	  continue;
	}
	io.drawLAN(newname,intfcptr[i]->typ, nx, ny);
	lan = lmgr.LAN_find(newname);           /* get ptr to new LAN */
	if (xcoord == 0) {
	  if (QUIETMODE) {
	    xcoord = nx + 80 + memNum * 4;
	    ycoord = ny + (memNum++) * 2;
	  }
	  else {
	    xcoord = nx + 80;               /* temporary positions for */
	    ycoord = ny + 30;               /*   new component */
	  }
	}
	lan->add_member(addr, intfcptr[i]->nr, xcoord, ycoord);/*add new comp*/
	mibvar.nr = 1;
	sprintf(mibvar.name[0],"ip_ipRoutingTable_ipRoute_ipRouteIfIndex_%s.0",
		intfcptr[i]->subnet+4);    /* skip the "LAN." */
	strcpy(cid, g->getcomm());
	errno = sendamsg(REQ, netaddr, cid, mibvar, response, 3, 6);
	if (errno == 0) {
	  k = response.val[0].value.intgr;
	  if (k < 0 || k > 100) k = 0;
	}
	else {
	  k = 0;
	}
	io.drawlink(newname, netaddr);
	lan->add_member(netaddr, k, g->getxcoord(), g->getycoord());
	g->addneighbor(newname);  /* add new LAN to neighbor's neighbor list */
	neighborptr = new addrlist;
	strcpy(neighborptr->addr, newname);  /* add LAN to the new */
	neighborptr->next = neighborlistptr; /*   component's */
	neighborlistptr = neighborptr;       /* neighbor list */
	sprintf(str, "Created new LAN between %s and %s for subnet %s\n",
		addr, netaddr, intfcptr[i]->subnet);
	io.printline(output1, str);
      }
      addrptr = addrptr->next;
    }
    while (addrptr != NULLPTR);
  }

/* Now check remote connections (interfaces with no direct connections) */
  notconnected = 0;
  for (i = 1; i <= nrOfIntfc; i++)   /* determine nr intfc empty */
    if (intfcptr[i]->nrOnNet == 0)
      notconnected++;
  if (notconnected > 0) {          /* if there are remotes, then continue */
    gfirst = lmgr.gtwy_getfirst();  /* save 1st, so can tell when gone */
    g = gfirst;                     /*  around list of gateways once */
  /* Look at each component in network to see if connected with hop = 0 */
    sprintf(st1, "ip_ipRoutingTable_ipRoute_ipRouteMetric1_");
    sprintf(st2, "ip_ipRoutingTable_ipRoute_ipRouteType_");
    sprintf(st3, "ip_ipRoutingTable_ipRoute_ipRouteIfIndex_");
    sprintf(st4, "ip_ipRoutingTable_ipRoute_ipRouteNextHop_");
    sprintf(st5, "ip_ipRoutingTable_ipRoute_ipRouteProto_");
    do {
      if (g == NULLPTR) break;                  /* no other gateways */
      strcpy(n, g->getaddr());
      mibvar.nr = 5;
      subnetmask = 4;
      errno = 1;
      while (subnetmask > 1 && errno != 0) {
	getsubnet(n, subnet, subnetmask--);
	sprintf(mibvar.name[0],"%s%s", st1, subnet);
	sprintf(mibvar.name[1],"%s%s", st2, subnet);
	sprintf(mibvar.name[2],"%s%s", st3, subnet);
	sprintf(mibvar.name[3],"%s%s", st4, subnet);
	sprintf(mibvar.name[4],"%s%s", st5, subnet);
	errno = sendamsg(REQ, addr, community, mibvar, response, 3, 6);
#ifdef DEBUG
	printf("errno: %d, sent to %s for info about %s\n", errno, addr, 
	       subnet);
#endif
      }
      if (errno == 0) {
	metric = response.val[0].value.intgr;
	rtype = response.val[1].value.intgr;
	intfc = response.val[2].value.intgr;
	if (intfc < 0 || intfc > nrOfIntfc) intfc = 0;
	strcpy(netaddr, response.asciival[3]);
	rproto = response.val[4].value.intgr;
#ifdef DEBUG
	sprintf(str, "for netaddr: %s, metric: %d, rtype: %d, intfc: %d,%d\n",
		n, metric, rtype, intfc, intfcptr[intfc]->nrOnNet);
	io.printline(output1, str);
#endif
	if (metric < 3 && intfcptr[intfc]->nrOnNet == 0 && 
	    strcmp(netaddr, "0.0.0.0") != 0) { 	  /* may be good connection */
	  strcpy(netname, "none");
	  k = lmgr.findNameAddr(netaddr,netalias,netname,cid,netsnmp,nettyp);
#ifdef DEBUG
	  if (k && strcmp(netaddr, netalias))
	    sprintf(str, "  nexthop is %s (%s)\n", netaddr, netalias);
	  else
	    sprintf(str, "  nexthop is %s\n", netaddr);
	  io.printline(output1, str);
#endif
	  if (!strcmp(netaddr, n)) {    /* good connection - it's next hop */
	    nrOfConnections++;
	    strcpy(cid, g->getcomm());
	    mibvar.nr = 1;
	    sprintf(mibvar.name[0], "%s%s", st3, addr);
	    errno = sendamsg(REQ, netaddr, cid, mibvar, response, 3, 6);
	    if (errno == 0)
	      k = response.val[0].value.intgr;   /* net component's intfc# */
	    else k = 0;                          /* this shouldn't happen */
#ifdef DEBUG
	    sprintf(str, "Interface %d for %s\n", k, netaddr);
	    io.printline(output1, str);
#endif
	    if (intfcptr[intfc]->typ > 1 && intfcptr[intfc]->typ < 18) {
	      if (strlen(intfcptr[intfc]->subnet) == 0) {
		sprintf(intfcptr[intfc]->subnet, "LAN.%s", netalias);
		j1 = strlen(intfcptr[intfc]->subnet);
		while (intfcptr[intfc]->subnet[j1] != '.' && j1 > 0)
		  intfcptr[intfc]->subnet[j1--] = '\0';
		intfcptr[intfc]->subnet[j1] = '\0';
	      }
	      if ((lan = lmgr.LAN_find(intfcptr[intfc]->subnet)) == NULLPTR) {
		nx = g->getxcoord() + 80;
		ny = g->getycoord();
		if (xcoord == 0) {
		  if (QUIETMODE) {
		    ny = ny + (nrOfIntfc - intfcptr[intfc]->nr/2) *40;
		    xcoord = nx + 80 + memNum * 4;
		    ycoord = ny + (memNum++) * 2;
		  }
		  else {
		    ny = ny + 30;
		    xcoord = nx + 80;
		    ycoord = ny + 30;
		  }
		}
		else ny = ny + 30;
		j1 = lmgr.LAN_add(intfcptr[intfc]->subnet, 
			  intfcptr[intfc]->typ, intfcptr[intfc]->speed, nx,ny);
		sprintf(str, "Added new LAN %s\n", intfcptr[intfc]->subnet);
		io.printline(NOPRESS, str);
		if (j1 != 0) {
		  io.printline(output1, "Unable to add new LAN - list full\n");
		  g = lmgr.gtwy_getnext();
		  continue;
		}
		io.drawLAN(intfcptr[intfc]->subnet,intfcptr[intfc]->typ,nx,ny);
		lan = lmgr.LAN_find(intfcptr[intfc]->subnet);
		lan->add_member(netaddr, k, g->getxcoord(), 
				g->getycoord());
		io.drawlink(netaddr, intfcptr[intfc]->subnet);
		g->addneighbor(intfcptr[intfc]->subnet);
	      }
	      else {                     /* LAN already exists */
		if (lan->is_member(addr)) {
		  g = lmgr.gtwy_getnext();
		  continue;
		}
		if (xcoord == 0) {
		  lan->getLANcoord(nx, ny);
		  xcoord = nx + 80 + (memNum++) * 4;
		  ycoord = ny + (memNum++) * 2;
		}
	      }
	      lan->add_member(addr, intfc, xcoord, ycoord); /* add node */
	      neighborptr = new addrlist;               /* and LAN to node */
	      strcpy(neighborptr->addr, intfcptr[intfc]->subnet);
	      neighborptr->next = neighborlistptr; /* new component's list */
	      neighborlistptr = neighborptr;
	      notconnected--;
	    } 
	    else {    /* it's remote */
	      if (xcoord == 0) {
		nx = g->getxcoord();
		ny = g->getycoord();
		xcoord = nx + memNum * 2;
		ycoord = ny + 60 + (memNum++) * 4;
	      }
	      if (k) 
		errno = lmgr.link_add(netaddr, addr, nx, ny, xcoord, ycoord,
				   intfcptr[k]->typ, intfcptr[k]->speed, rtype,
				   rproto, k, intfc);
	      else {
		errno = lmgr.link_add(netaddr, addr, nx, ny, xcoord, ycoord,
				      22, 1, rtype, rproto, 0, intfc);
		io.printline(output1,"WARNING - link has incomplete info\n");
	      }
	      if (errno != 0) {
		io.printline(output1, "Unable to add new link - list full\n");
		nrOfConnections--;
	      }
	      else {
		sprintf(str,"Created link between %s and %s\n",netaddr,addr);
		io.printline(output1, str);
		intfcptr[intfc]->nrOnNet++;
		notconnected--;
		g->addneighbor(addr);      /* add component to net neighbor */
		neighborptr = new addrlist;
		strcpy(neighborptr->addr, netaddr);  /* add net neighbor to */
		neighborptr->next = neighborlistptr; /* new component's list */
		neighborlistptr = neighborptr;
	      }
	    }
	  }
#ifdef DEBUG
	  else {
	    sprintf(str, "%s not adjacent\n", netaddr);
	    io.printline(output1, str);
	  }
#endif
	}
	else {
#ifdef DEBUG
	  sprintf(str, "%s not adjacent (%d hops, %d rtype)\n", netaddr,
		  metric, rtype);
	  io.printline(output1, str);
#endif
	  if (intfcptr[intfc]->nrOnNet>1 && rtype == 4) {
	    io.printline(output1,"ACTUAL NETWORK ERROR: ");
	    sprintf(str, "%s shows LAN interface %d to be remote route.\n",
		    addr, intfc);
	    io.printline(output1, str);
	  }
	}
      }
      else 
	if (errno < 0) {
	  sprintf(str, "***errno: %d for netaddr %s\n", errno, netaddr);
	  io.printline(output1, str);
	}
      g = lmgr.gtwy_getnext();
    }
    while (g != gfirst);
  }

  /* If there are still unconnected interfaces, look at Nexthops to determine
     what those interfaces are connnected to. */ 
  if (notconnected > 0) {
    k = 0;
    mibvar.nr = 4;
    mibvar1.nr = 4;
    sprintf(st1, "ip_ipRoutingTable_ipRoute_ipRouteType_");
    sprintf(st2, "ip_ipRoutingTable_ipRoute_ipRouteIfIndex_");
    sprintf(st3, "ip_ipRoutingTable_ipRoute_ipRouteNextHop_");
    sprintf(st4, "ip_ipRoutingTable_ipRoute_ipRouteProto_");
    sprintf(mibvar.name[0],"%s", st1);
    sprintf(mibvar.name[1],"%s", st2);
    sprintf(mibvar.name[2],"%s", st3);
    sprintf(mibvar.name[3],"%s", st4);
    sprintf(mibvar1.name[0],"%s", st1);
    sprintf(mibvar1.name[1],"%s", st2);
    sprintf(mibvar1.name[2],"%s", st3);
    sprintf(mibvar1.name[3],"%s", st4);
    do {
      errno = sendamsg(NXT, addr, community, mibvar, response, 3, 6);
      k++;
      if (errno != 0) continue;
      sprintf(mibvar.name[0], "%s%s", st1, response.vars.name[2]+41);
      sprintf(mibvar.name[1],  "%s%s", st2, response.vars.name[2]+41);
      sprintf(mibvar.name[2],  "%s%s", st3, response.vars.name[2]+41);
      sprintf(mibvar.name[3],  "%s%s", st4, response.vars.name[2]+41);
      strcpy(netaddr, response.asciival[2]);
      if (strcmp(netaddr, "0.0.0.0") == 0) continue;
      intfc = response.val[1].value.intgr;
      if (intfc < 1 || intfc > nrOfIntfc) continue;
      rtype = response.val[0].value.intgr;
      rproto = response.val[3].value.intgr;
/*%%%%printf("**Checking intfc %d for %s (%d on net)\n", intfc, netaddr,
	       intfcptr[intfc]->nrOnNet); */
/*%%%%printf("   %d connections still unresolved\n", notconnected); */
      if (intfcptr[intfc]->nrOnNet == 0) {
	intfcptr[intfc]->nrOnNet++;
	notconnected--;
      }
      strcpy(netname, "none");
      addrfound = lmgr.findNameAddr(netaddr, netalias, netname, cid, netsnmp,
				    nettyp);
      if ((g = lmgr.gtwy_find(netaddr)) == NULLPTR) continue; /*not in net*/
      if(!notonneighborlist(neighborlistptr, netaddr)) continue;/*is neighbor*/
      if (intfcptr[intfc]->typ > 1 && intfcptr[intfc]->typ < 18) { /* a LAN */
	if (strlen(intfcptr[intfc]->subnet) == 0) {
	  sprintf(intfcptr[intfc]->subnet, "LAN.%s", netalias);
	  j1 = strlen(intfcptr[intfc]->subnet);
	  while (intfcptr[intfc]->subnet[j1] != '.' && j1 > 0)
	    intfcptr[intfc]->subnet[j1--] = '\0';
	  intfcptr[intfc]->subnet[j1] = '\0';
	}
	if (!notonneighborlist(neighborlistptr, intfcptr[intfc]->subnet))
	  continue;         /* LAN is already a neighbor - so skip */
      }
#ifdef DEBUG
      sprintf(str,"Found %s uses intfc %d to %s\n", addr, intfc, netaddr);
      io.printline(NOPRESS, str);
#endif

      k1 = 0;                              /* counter for nr messages sent */
      netintfc = 0;                        /* intfc nr for net member */
      /* Now check network member to get intfc nr back to new node */
      do {
	errno = sendamsg(NXT, netaddr, cid, mibvar1, response, 3, 6);
	k1++;
	if (errno != 0) continue;
	sprintf(mibvar1.name[0], "%s%s", st1, response.vars.name[2]+41);
	sprintf(mibvar1.name[1],  "%s%s", st2, response.vars.name[2]+41);
	sprintf(mibvar1.name[2],  "%s%s", st3, response.vars.name[2]+41);
	sprintf(mibvar1.name[3],  "%s%s", st4, response.vars.name[2]+41);
	strcpy(n, response.asciival[2]);
	if (strcmp(n, "0.0.0.0") == 0) continue;
	i = response.val[1].value.intgr;
	if (i < 1 || i > 30) continue;
	strcpy(netname, "none");
	addrfound = lmgr.findNameAddr(n, nalias, netname, ncid, netsnmp,
				      nettyp);
	if (strcmp(n,addr) == 0) {
	  netintfc = i;
#ifdef DEBUG
	  sprintf(str, "Found %s uses intfc %d for %s\n", netaddr, netintfc,
		  addr);
	  io.printline(NOPRESS, str);
#endif
	  break;
	}
      }
      while (k1 < 50);         /* quit if haven't found intfc in 50 msgs */
      /* Add component - if on LAN, add new LAN if not already in net */
      if (intfcptr[intfc]->typ > 1 && intfcptr[intfc]->typ < 18) { /* a LAN */
/*%%%%%printf("**Looking for LAN %s\n", intfcptr[intfc]->subnet); */
	if ((lan = lmgr.LAN_find(intfcptr[intfc]->subnet)) == NULLPTR) {
	  nx = g->getxcoord() + 80;
	  ny = g->getycoord();
	  if (xcoord == 0) {
	    if (QUIETMODE) {
	      ny = ny + (nrOfIntfc - intfcptr[intfc]->nr/2) * 40;
	      xcoord = nx + 80 + memNum * 4;
	      ycoord = ny + (memNum++) * 2;
	    }
	    else {
	      ny = ny + 30;
	      xcoord = nx + 80;
	      ycoord = yx + 30;
	    }
	  }
	  else ny = ny + 30;
	  j1 = lmgr.LAN_add(intfcptr[intfc]->subnet, intfcptr[intfc]->typ,
			    intfcptr[intfc]->speed, nx, ny);
	  sprintf(str, "Added new LAN %s\n", intfcptr[intfc]->subnet);
	  io.printline(NOPRESS, str);
	  if (j1 != 0) {
	    io.printline(output1, "Unable to add new LAN - list full\n");
	    continue;
	  }
	  io.drawLAN(intfcptr[intfc]->subnet, intfcptr[intfc]->typ, nx, ny);
	  lan = lmgr.LAN_find(intfcptr[intfc]->subnet);
	  lan->add_member(netaddr, netintfc, g->getxcoord(), g->getycoord());
	  io.drawlink(netaddr, intfcptr[intfc]->subnet);
	  g->addneighbor(intfcptr[intfc]->subnet);
	}
	else {                     /* LAN already exists */
	  if (lan->is_member(addr)) continue;
	  if (xcoord == 0) {
	    lan->getLANcoord(nx, ny);
	    if (QUIETMODE) {
	      xcoord = nx + 80 + memNum * 4;
	      ycoord = ny + (memNum++) * 2;
	    }
	    else {
	      xcoord = nx + 80;
	      ycoord = ny + 30;
	    }
	  }
	}
	lan->add_member(addr, intfc, xcoord, ycoord); /* add new node to LAN */
	neighborptr = new addrlist;                   /* and LAN to node */
	strcpy(neighborptr->addr, intfcptr[intfc]->subnet);
	neighborptr->next = neighborlistptr; /* new component's list */
	neighborlistptr = neighborptr;
	nrOfConnections++;
      } 
      else {                       /* otherwise it's remote - add link */
	g->addneighbor(addr);      /* put new node on list of net member */
	neighborptr = new addrlist;
	strcpy(neighborptr->addr, netaddr);  /* add net neighbor to */
	neighborptr->next = neighborlistptr; /* new component's list */
	neighborlistptr = neighborptr;
	if (xcoord == 0) {
	  nx = g->getxcoord();
	  ny = g->getycoord();
	  xcoord = nx + memNum * 2;
	  ycoord = ny + 60 + (memNum++) * 4;
	}
	sprintf(str, "Added new link from %s to %s\n", addr, netaddr);
	io.printline(output1, str);
	errno = lmgr.link_add(addr, netaddr, xcoord, ycoord, nx, ny,
			      intfcptr[intfc]->typ, intfcptr[intfc]->speed,
			      rtype, rproto, intfc, netintfc);
	nrOfConnections++;
      }
    }
    while (k < 150 && notconnected > 0);
  }

  /* Don't add if no connections */
  if (nrOfConnections == 0 && lmgr.nrGateways() > 0) {
    sprintf(str, "Unable to add %s - no connections to existing network\n",
	    addr);
    io.printline(output2, str);
    return -4;
  }

/* Print and save the new component */
  j = intfcptr[1]->typ;                       /* determine if all connected */
  for (i = 2; i <= nrOfIntfc; i++)           /* nets are same type */
    if (intfcptr[i]->typ != j) {
      j = -99;
      break;
    }
  if (j < 0) j = GATEWAY;                    /* >1 type means a gateway */
  else
    if (nrOfIntfc > 2) j = ROUTER;           /* >2 of same type is router */
    else 
      if (j < 2) j = HOST;                   /* 1 intfc means host */
      else {                                 /* 2 intfc - same subnet? */
	if (strcmp(intfcptr[1]->subnet, intfcptr[2]->subnet) == 0) j = HOST;
	else j = ROUTER;
      }
  k = lmgr.saveNameAddr(addr, alias, name, community, YES, j);
  k = lmgr.gtwy_add(addr, name, community, YES, YES, j);
  if (k != 0)
    io.printline(output1, "UNABLE to add new component - list full\n");
  g = lmgr.gtwy_find(addr);                  /* get ptr to new component */
  if (xcoord == 0) xcoord = 149;             /* default */
  if (ycoord == 0) ycoord = 149;             /*   coordinates */
  g->setcoords(xcoord, ycoord);              /* set xy coords */
  io.drawnode(addr,name,YES,YES,xcoord,ycoord,j);
  io.printline(output1, "\nAdded new component:\n");
  sprintf(str, "  Address:   %s\n", addr);
  io.printline(output1, str);
  sprintf(str, "  Name:      %s\n", name);
  io.printline(output1, str);
  sprintf(str, "  Community: %s\n", community);
  io.printline(output1, str);
  sprintf(str, "  Xcoord:    %d\n", xcoord);
  io.printline(output1, str);
  sprintf(str, "  Ycoord:    %d\n", ycoord);
  io.printline(output1, str);
  switch(j) {
  case GATEWAY: io.printline(output1, "  Type:      gateway\n");
                break;
  case ROUTER: io.printline(output1, "  Type:      router\n");
               break;
  case HOST: io.printline(output1, "  Type:      host\n");
             break;
  default: break;
  }
  while (neighborlistptr != NULLPTR) {       /* add neighbors */
    g->addneighbor(neighborlistptr->addr);
    sprintf(str, "  Neighbor:  %s\n", neighborlistptr->addr);
    io.printline(output1, str);
    io.drawlink(addr,neighborlistptr->addr);
    neighborptr = neighborlistptr;
    neighborlistptr = neighborlistptr->next;
    delete neighborptr;                      /* clean up after thyself */
  }
  for (i = 1; i <= nrOfIntfc; i++) {         /* more clean-up */
    while (intfcptr[i]->adjptr != NULLPTR) {
      addrptr = intfcptr[i]->adjptr;          /* delete intf connection list */
      intfcptr[i]->adjptr = intfcptr[i]->adjptr->next;
      delete addrptr;
    }
    delete intfcptr[i];                      /* delete interface itself */
  }
  io.printline(output2, "\n");
  return 0;
}

void actmgr::map(mstring str, listmgr& lmgr, iomgr& io) {
  ipaddr addr, origaddr, alias, netaddr, netalias;
  nametype name, netname, newname;
  commid cid, netcid;
  mstring data, str0, str1, str2, str3, str4, str5, str6;
  gateway *g_map, *g;
  LAN *lan;
  link *l;
  struct interface intfc;
  vartable mibvar, mibvar1;
  responseList response;
  addrlist *addrlistptr = NULLPTR, *addrptr;
  int snmponly = TRUE, i, j, snmp = 0, typ = 0, addrinworld, nrOfIntfc, errno;
  int nettyp, netsnmp, k, k1, added, hasremote = 0, netx, nety;
  int rtype, rintfc, rproto, addrfound, xcoord, ycoord, lanx, lany;
  unsigned long speed;
  sscanf(str, "%s%s%s%s", addr, name, str0, data);
  if (str0[0] == 'c') return;  /* confirm not implemented yet */
  if (strncmp(data, "any", 3) == 0) snmponly = FALSE;
  strcpy(origaddr, addr);
  strcpy(cid, "public");
  io.openDialog(24);
  if ((addrinworld = lmgr.findNameAddr(addr,alias,name,cid,snmp,typ)) == 0) {
    sprintf(str, "Cannot map - unable to identify %s\n", origaddr);
    io.printline(PRESS, str);
    return;
  }
  if (strncmp(origaddr, "none", 4) == 0) strcpy(origaddr, addr);
  if (addrinworld > 2) addrinworld = 0;  /* add new info to lclworld later */
  if ((g_map = lmgr.gtwy_find(addr)) == NULLPTR) {
    sprintf(str, "Cannot map - %s not in current network\n", origaddr);
    io.printline(PRESS, str);
    return;
  }
  sprintf(str,"Mapping: %s (%s)\n\n", addr, name);
  io.printline(NOPRESS, str);
  mibvar.nr = 1;
  strcpy(mibvar.name[0],"interfaces_ifNumber_0");
  errno = sendamsg(REQ, addr, cid, mibvar, response, 3, 6);
  if (errno) {
    sprintf(str, "Cannot map - %s not responding to SNMP message\n",origaddr);
    io.printline(PRESS, str);
    return;
  }
  else {
    nrOfIntfc = response.val[0].value.intgr;
    sprintf(str, "%s has %d interfaces - Checking LAN connections first...\n",
	    addr, nrOfIntfc);
    io.printline(NOPRESS, str);
  }
  if (nrOfIntfc > 49) nrOfIntfc = 49;
  intfc.nr = 0;             /* 0 used for default interface nr later */
  for (i = 1; i <= nrOfIntfc; i++) {
    mibvar.nr = 4;
    sprintf(mibvar.name[0], "interfaces_ifTable_ifEntry_ifIndex_%d", i);
    sprintf(mibvar.name[1], "interfaces_ifTable_ifEntry_ifDescr_%d", i);
    sprintf(mibvar.name[2], "interfaces_ifTable_ifEntry_ifType_%d", i);
    sprintf(mibvar.name[3], "interfaces_ifTable_ifEntry_ifSpeed_%d", i);
    j = sendamsg(REQ, addr, cid, mibvar, response, 3, 6);
    if (j == 0) {
      intfc.nr = response.val[0].value.intgr;
      strcpy(intfc.descr, response.asciival[1]);
      intfc.typ = response.val[2].value.intgr;
      intfc.speed = response.val[3].value.cntr;
    }
    else {
      intfc.nr = i;
      strcpy(intfc.descr, "unknown");
      intfc.typ = 1;
      intfc.speed = 0;
    }
    sprintf(str, "Interface:   %d\n", intfc.nr);
    io.printline(NOPRESS, str);
    sprintf(str, "  Descr:     %s\n", intfc.descr);
    io.printline(NOPRESS, str);
    sprintf(str, "  Type code: %d\n", intfc.typ);
    io.printline(NOPRESS, str);
    sprintf(str, "  Speed:     %lu\n", intfc.speed);
    io.printline(NOPRESS, str);
    intfc.subnet[0] = '\0';
    mibvar.nr = 1;                   /* now look for adjacent components */
    response.vars.nr = 0;            /*   on this interface */
    sprintf(data, "at_atTable_atEntry_atNetAddress_%d",i);
    strcpy(mibvar.name[0], data);
    memNum = intfc.nrOnNet = 0;          /* total nr on interface */
    intfc.subnet[0] = '\0';
    j = 1;                             /* =0 signals end of interface data */
    do {
      errno = sendamsg(NXT, addr, cid, mibvar, response, 3, 6);
      if (errno == 0) {
	strcpy(mibvar.name[0], response.vars.name[0]);  /* setup for next */
	if (strncmp(mibvar.name[0], data, strlen(data)) == 0) {
	  intfc.nrOnNet++;       /* another comp. on same interface */
	  strcpy(netaddr, response.asciival[0]);  /* get addr returned */
	  strcpy(netname, "none");
	  strcpy(netcid, cid);
	  netsnmp = UNK;
	  nettyp = UNK;
	  k = lmgr.findNameAddr(netaddr, netalias, netname, netcid,
					netsnmp, nettyp);
	  if (k < 1 || k > 2) strcpy(netcid, cid); /* use mapper if cid unk */
	  if (strlen(intfc.subnet) == 0) {
	    sprintf(intfc.subnet, "LAN.%s",netalias);/* the subnet */
	    k = strlen(intfc.subnet);             /* identifier for */
	    while (intfc.subnet[k] != '.' && k > 0)/* the subnet */
	      intfc.subnet[k--] = '\0';           /* attached to */
	    intfc.subnet[k] = '\0';               /* this interface */
	  }
	  sprintf(str,"%s:   Checking %s (%s)\n",intfc.subnet,netaddr,netname);
	  io.printline(NOPRESS, str);
	  if ((lan = lmgr.LAN_find(intfc.subnet)) == NULLPTR) {
	    strcpy(newname, intfc.subnet);
	    added = lmgr.LAN_add(newname, intfc.typ, intfc.speed, 
				 g_map->getxcoord()+ 80, g_map->getycoord()+ 
				 (intfc.nr - nrOfIntfc/2) * 40);
	    io.drawLAN(newname, intfc.typ, g_map->getxcoord() + 80,
		       g_map->getycoord() + (intfc.nr - nrOfIntfc/2) * 40);
	    g_map->addneighbor(newname);
	    io.drawlink(newname, addr);
	    lan = lmgr.LAN_find(newname);
	    lan->add_member(addr, i, g_map->getxcoord(), g_map->getycoord());
	    sprintf(str, "Created %s\n", intfc.subnet);
	    io.printline(NOPRESS, str);
	  }
	  if ((g = lmgr.gtwy_find(netaddr)) == NULLPTR) { /* not in network */
	    sprintf(str,"%s %s %s", netaddr, netname, netcid);
	    strcpy(str1, str);
	    added = addcomponent(QUIETMODE, str, lmgr, io);
#ifdef DEBUG
	    sprintf(str, "Results from adding: %s: %d\n", str1, added);
	    io.printline(NOPRESS, str);
#endif
	    switch (added) {
	    case 0:  /* added okay */
	      sprintf(str, "%s: %s added (new component)\n",
		      intfc.subnet, netaddr);
	      break;
	    case -1:  /* no snmp */
	      if (snmponly)
		sprintf(str, "%s: %s not added - no snmp\n",
			intfc.subnet, netaddr);
	      else {
		sprintf(str, "%s: %s added (no snmp)\n", intfc.subnet,netaddr);
		lmgr.gtwy_add(netaddr, netname, netcid, UNK, NO, UNK);
		g = lmgr.gtwy_find(netaddr);
		g->addneighbor(intfc.subnet);
		lan->getLANcoord(netx, nety);
		netx = netx + 80 + memNum * 4;
		nety = nety + (memNum++) * 2;
		g->setcoords(netx, nety);
		io.drawnode(netaddr, netname, UNK, NO, netx, nety, UNK);
		io.drawlink(intfc.subnet, netaddr);
		lan->add_member(netaddr, 0, netx, nety);
	      }
	      break;
	    case -2:   /* no such addr */
	      sprintf(str, "%s: %s not added - addr not recognized\n",
		      intfc.subnet, netaddr);
	      break;
	    case -5:   /* msg response failure */
	      sprintf(str, "%s: %s added but info incomplete\n",
		      intfc.subnet, netaddr);
	      lmgr.gtwy_add(netaddr, netname, netcid, UNK, NO, UNK);
	      g = lmgr.gtwy_find(netaddr);
	      g->addneighbor(intfc.subnet);
	      lan->getLANcoord(netx, nety);
	      netx = netx + 80 + memNum * 4;
	      nety = nety + (memNum++) * 2;
	      g->setcoords(netx, nety);
	      io.drawnode(netaddr, netname, UNK, NO, netx, nety, UNK);
	      io.drawlink(intfc.subnet, netaddr);
	      lan->add_member(netaddr, 0, netx, nety);
	      break;
	    case -4:;   /* no connections */
	      sprintf(str, "%s: %s added but info incomplete\n",
		      intfc.subnet, netaddr);
	      lmgr.gtwy_add(netaddr, netname, netcid, YES, YES, UNK);
	      g = lmgr.gtwy_find(netaddr);
	      g->addneighbor(intfc.subnet);
	      lan->getLANcoord(netx, nety);
	      netx = netx + 80 + memNum * 4;
	      nety = nety + (memNum++) * 2;
	      g->setcoords(netx, nety);
	      io.drawnode(netaddr, netname, YES, YES, netx, nety, UNK);
	      io.drawlink(intfc.subnet, netaddr);
	      lan->add_member(netaddr, 0, netx, nety);
	      break;
	    case -3:;  /* already in network */
	    default:
	      break;
	    }
	    io.printline(NOPRESS, str);
	  }
	}
	else {
	  if (intfc.nrOnNet == 0) hasremote++;
	  j = 0;               /* no longer same interface */
	}
      }
    }  /* end do - look for another neighbor on this interface? */
    while (errno == 0 && j == 1);
  }

  /* Now check for remote neighbors with hop of 1 */
  memNum = 0;
  sprintf(str1, "\nNow checking for remote neighbors...\n");
  io.printline(NOPRESS, str1);
  strcpy(str1, "ip_ipRoutingTable_ipRoute_ipRouteNextHop_");
  strcpy(str2, "ip_ipRoutingTable_ipRoute_ipRouteIfIndex_");
  strcpy(str3, "ip_ipRoutingTable_ipRoute_ipRouteType_");
  strcpy(str4, "ip_ipRoutingTable_ipRoute_ipRouteProto_");
  strcpy(str5, "interfaces_ifTable_ifEntry_ifSpeed_");
  strcpy(str6, "interfaces_ifTable_ifEntry_ifType_");
  mibvar.nr = 4;
  strcpy(mibvar.name[0], str1);
  strcpy(mibvar.name[1], str2);
  strcpy(mibvar.name[2], str3);
  strcpy(mibvar.name[3], str4);
  mibvar1.nr = 2;
  strcpy(mibvar1.name[0], str5);
  strcpy(mibvar1.name[1], str6);
  for(k1 = 0; k1 < 100; k1++) {
    if (hasremote == 0) break;
/*%%%% printf("hasremote: %d, k1: %d, %s\n",hasremote,k1,mibvar.name[0]); */
    errno = sendamsg(NXT, addr, cid, mibvar, response, 3, 4);
    if (errno != 0) continue;
    sprintf(mibvar.name[0], "%s%s", str1, response.vars.name[0]+41);
    sprintf(mibvar.name[1], "%s%s", str2, response.vars.name[0]+41);
    sprintf(mibvar.name[2], "%s%s", str3, response.vars.name[0]+41);
    sprintf(mibvar.name[3], "%s%s", str4, response.vars.name[0]+41);
    if (strncmp(mibvar.name[0], str1, strlen(str1)) != 0)
      break;                          /* quit when no longer NextHop data */
    strcpy(netaddr, response.asciival[0]);  /* the network response */
    if (strcmp(netaddr, "0.0.0.0") == 0) continue;
    rintfc = response.val[1].value.intgr;
    if (rintfc < 0 || rintfc > 30) continue;
    rtype = response.val[2].value.intgr;
    rproto = response.val[3].value.intgr;
/*%%%%printf("  remote nexthop: %s, intfc: %d\n", netaddr, rintfc); */
    i = strlen(netaddr-1);                  /* the last char of it */
    while (netaddr[i] != '.' && i > 0) i--; /* find the last '.' */
    j = 0;
    if (i > 0) {
      j = atoi(netaddr + i + 1);      /* get the last nr in the addr */
      if (j < 0 || j > 254) j = 0;    /* such as 0.0.0.12 -> 12 */
    }
    if (j <= 0) continue;             /* = 0 means a subnet - no good */
    strcpy(netname, "none");
    strcpy(netcid, cid);
    netsnmp = UNK;
    nettyp = UNK;
    addrfound = lmgr.findNameAddr(netaddr, netalias, netname, netcid,
				  netsnmp, nettyp);
    if (addrfound && notonneighborlist(addrlistptr, netaddr) &&
	((g = lmgr.gtwy_find(netaddr)) == NULLPTR)) {
      addrptr = new addrlist;
      strcpy(addrptr->addr, netaddr);  /* If valid addr and not already */
      addrptr->next = addrlistptr;     /*  in the network, put at the */
      addrlistptr = addrptr;           /*  head of list of others */
      sprintf(str, "%s may be adjacent\n", addrptr->addr);
      io.printline(NOPRESS, str);
      sprintf(str, "%s %s %s", netaddr, "none", netcid);
      added = addcomponent(QUIETMODE, str, lmgr, io);
      str[0] = '\0';
      switch (added) {
      case 1: break;  /* not a valid entry */
      case 0:         /* added okay */
	sprintf(str, "remote: %s added (new component)\n", netaddr);
	io.printline(NOPRESS, str);
	memNum++;
	g = lmgr.gtwy_find(netaddr);
	if ((l = lmgr.link_find(addr, netaddr)) != NULLPTR) {  
	  sprintf(str,"remote: Created link between %s and %s\n",addr,netaddr);
	  io.printline(NOPRESS, str);
	  hasremote--;                  /* if link added between mapper and */
	  break;                        /*   adjacent node, then okay */
	}
	sprintf(intfc.subnet, "LAN.%s", netalias); /* get LAN name */
	k = strlen(intfc.subnet);
	while (intfc.subnet[k] != '.' && k > 0)
	  intfc.subnet[k--] = '\0';
	intfc.subnet[k] = '\0';
	if ((lan = lmgr.LAN_find(intfc.subnet)) != NULLPTR)/* if mapper and */
	  if (lan->is_member(addr)) {             /* adjacent node on a LAN */
	    sprintf(str,"remote: %s added to %s\n",addr, intfc.subnet);
	    io.printline(NOPRESS, str);
	    break;                                /* then all is okay */
	  }
      case -1:   /* no snmp */                    /* otherwise skip this */
	if (snmponly && added == -1) {            /* and force add below */
	  sprintf(str, "remote: %s not added - no snmp\n", netaddr);
	  io.printline(NOPRESS, str);
	  break;
	}
      case -5:;  /* msg failure */
      case -4:;  /* addcomponent found no connection */
	sprintf(intfc.subnet, "LAN.%s",netalias); /* the subnet */
	k = strlen(intfc.subnet);                 /* identifier for */
	while (intfc.subnet[k] != '.' && k > 0)   /* the subnet */
	  intfc.subnet[k--] = '\0';               /* attached to */
	intfc.subnet[k] = '\0';                   /* this interface */
	sprintf(mibvar1.name[0], "%s%d", str5, rintfc);
	sprintf(mibvar1.name[1], "%s%d", str6, rintfc);
	errno = sendamsg(REQ, addr, cid, mibvar1, response, 3, 6);
	if (errno == 0) {
	  speed = response.val[0].value.cntr; /* interface speed */
	  j = response.val[1].value.intgr;     /* interface type */
	}
	else {
	  j = 1;
	  speed = 0;
	}
	if (added == -1) netsnmp = UNK;
	else netsnmp = YES;
	if (j > 1 && j < 18) {    /* it's a LAN */
	  if ((lan = lmgr.LAN_find(intfc.subnet)) == NULLPTR) {
	    strcpy(newname, intfc.subnet);    /* add the new LAN */
	    lanx = g_map->getxcoord() + 80;
	    lany = g_map->getycoord() + (rintfc - nrOfIntfc/2 * 40);
	    added = lmgr.LAN_add(newname, j, speed, lanx, lany);
	    sprintf(str,"remote: Created %s and added %s\n", newname, addr);
	    io.printline(NOPRESS, str);
	    io.drawLAN(newname, intfc.typ, lanx, lany);
	    g_map->addneighbor(newname);        /* add LAN to Nexthop's list */
	    lan = lmgr.LAN_find(newname);
	    lan->add_member(addr, 0, g_map->getxcoord(), g_map->getycoord());
	    io.drawlink(newname, addr);
	    hasremote--;
	  }
	  if (lmgr.gtwy_find(netaddr) == NULLPTR) {
	    i = lmgr.gtwy_add(netaddr, netname, netcid, UNK, netsnmp, UNK);
	    sprintf(str,"remote: %s added to network (data missing)\n", 
		    netaddr);
	    io.printline(NOPRESS, str);
	  }
	  g = lmgr.gtwy_find(netaddr);
	  g->addneighbor(newname);
	  lan = lmgr.LAN_find(newname);
	  lan->getLANcoord(lanx, lany);
	  lan->add_member(netaddr, i, lanx + 80 + memNum *4, lany + memNum *2);
	  g->setcoords(lanx + 80 + memNum * 4, lany + (memNum++) * 2);
	  io.drawlink(newname, netaddr);
	  sprintf(str, "remote: Added %s to %s\n", netaddr, newname);
	  io.printline(NOPRESS, str);
	}
	else {    /* add as remote link */
	  if (lmgr.gtwy_find(netaddr) == NULLPTR) {
	    sprintf(str, "remote: %s added (data incomplete)\n", netaddr);
	    io.printline(NOPRESS, str);
	    lmgr.gtwy_add(netaddr, netname, netcid, UNK, netsnmp, UNK);
	  }
	  g = lmgr.gtwy_find(netaddr);
	  g->addneighbor(addr);
	  xcoord = g_map->getxcoord();
	  ycoord = g_map->getycoord();
	  netx = xcoord + memNum *2;
	  nety = ycoord + 60 + (memNum++) * 4;
	  g->setcoords(netx, nety);
	  g_map->addneighbor(netaddr);
	  io.drawnode(netaddr, netname, UNK, netsnmp, netx, nety, UNK);
	  i = lmgr.link_add(addr, netaddr, xcoord, netx, ycoord, nety, j,
			    speed, rtype, rproto, rintfc, 0);
	  io.drawlink(netaddr, addr);
	  sprintf(str,"remote: Created link between %s and %s\n",netaddr,addr);
	  io.printline(NOPRESS, str);
	  hasremote--;
	}
	break;
      case -2:  /* no such addr */
	sprintf(str, "remote: %s not added - addr not recognized\n", netaddr);
	io.printline(NOPRESS, str);
	break;
      case -3:;  /* already in network */
      default:
	break;
      }
    }
  }
  io.printline(PRESS, "Mapping completed...\n");
}

void actmgr::trace(mstring str, listmgr& lmgr, iomgr& io) {
  ipaddr addr, addrStrt, aliasStrt, addrEnd, aliasEnd, addrNxt, aliasNxt;
  ipaddr subnet, addrprior;
  nametype nameStrt, nameEnd, nameNxt;
  commid cidStrt, cidEnd, cidNxt, cid;
  mstring str0, str1, str2, str3, typstr, protstr;
  gateway *g;
  int addrStrtInWorld, addrEndInWorld, addrNxtInWorld;
  int snmpStrt, typStrt, snmp, typ, rtyp, rproto;
  int subnetmask, errno, hopnr = 0, nr, delay, i, add = 0, metric;
  vartable mibvar;
  responseList response;
  sscanf(str, "%s%s%s%s%s%s",addrStrt,nameStrt,addrEnd,nameEnd,cidEnd,str0);
  add = atoi(str0);
  strcpy(cid, "public");
  strcpy(addr, addrStrt);
  io.openDialog(24);
  if ((addrStrtInWorld = lmgr.findNameAddr(addrStrt, aliasStrt, nameStrt,
					   cidStrt, snmpStrt, typStrt)) == 0) {
    sprintf(str, "Cannot trace - unable to identify %s\n", addr);
    io.printline(PRESS, str);
    return;
  }
  if (strncmp(addr, "none", 4) == 0) strcpy(addr, addrStrt);
  if (addrStrtInWorld > 2) addrStrtInWorld = 0;  /* add to lclworld later */
  if ((g = lmgr.gtwy_find(addrStrt)) == NULLPTR) {
    sprintf(str, "Cannot trace - %s not in current network\n", addr);
    io.printline(PRESS, str);
    return;
  }
  addrEndInWorld = lmgr.findNameAddr(addrEnd,aliasEnd,nameEnd,cid,snmp,typ);
  if (addrEndInWorld == 1 || addrEndInWorld ==2) strcpy(cidEnd, cid);
  else addrEndInWorld = 0;
  if (strncmp(addr, "none", 4) == 0) strcpy(addr, addrStrt);
  if (addrStrtInWorld > 2) addrStrtInWorld = 0;  /* add to lclworld later */
  sprintf(str, "Start at:  %s (%s)\n", addrStrt, nameStrt);
  io.printline(NOPRESS, str);
  sprintf(str, "Trace to:  %s (%s)\n\n", addrEnd, nameEnd);
  io.printline(NOPRESS, str);
  strcpy(addrNxt, addrStrt);
  strcpy(nameNxt, nameStrt);
  strcpy(cidNxt, cidStrt);
  snmp = snmpStrt;
  addrNxtInWorld = addrStrtInWorld;
  strcpy(str0, "ip_ipRoutingTable_ipRoute_ipRouteNextHop_");
  strcpy(str1, "ip_ipRoutingTable_ipRoute_ipRouteType_");
  strcpy(str2, "ip_ipRoutingTable_ipRoute_ipRouteProto_");
  strcpy(str3, "ip_ipRoutingTable_ipRoute_ipRouteMetric1_");
  mibvar.nr = 4;
  typstr[0] = '\0';
  protstr[0] = '\0';
  strcpy(str, "     Nr Address          Name                       ");
  strcat(str, "   Connect  Protocol  Left\n\n");
  io.printline(NOPRESS, str);
  do {
    errno = 1;
    subnetmask = 3;                          /* use 1st 3 nr's in addr */
    while (subnetmask >= 0 && errno != 0) {
      getsubnet(addrEnd, subnet, subnetmask--);
      sprintf(mibvar.name[0], "%s%s", str0, subnet);
      sprintf(mibvar.name[1], "%s%s", str1, subnet);
      sprintf(mibvar.name[2], "%s%s", str2, subnet);
      sprintf(mibvar.name[3], "%s%s", str3, subnet);
      delay = 6;
      if (snmp == YES) nr = 3;
      else nr = 2;
      errno = sendamsg(REQ, addrNxt, cidNxt, mibvar, response, nr, delay);
      if (errno < 0 && strcmp(cidNxt, cidEnd) != 0) {
	sprintf(str, "No snmp response for %s, %s - trying community %s...\n",
		addrNxt, cidNxt, cidEnd);
	io.printline(LOG, str);
	strcpy(cidNxt, cidEnd);
	errno = sendamsg(REQ, addrNxt, cidNxt, mibvar, response, nr, delay);
      }
#ifdef DEBUG
      sprintf(str, "    Queried: %s, error: %d, %s\n", addrNxt, errno, 
	      mibvar.name[0]);
      io.printline(NOPRESS, str);
#endif
      if (errno < 0) {
	sprintf(str, "Hop %2d: %-16.16s %-30.30s trace ended (no snmp)\n\n",
		hopnr, addrNxt, nameNxt);
	io.printline(PRESS, str);
/*	if (add && (g = lmgr.gtwy_find(addrEnd)) == NULLPTR) {
	  sprintf(str, "%s %s %s", addrEnd, nameEnd, cidEnd);
	  i = addcomponent(QUIETMODE, str, lmgr, io);
	} */
	return;
      }
      if (errno > 0) {
	if (subnetmask < 0) {
	  sprintf(str,"Hop %2d: %-16.16s %-30.30s trace ended (unk route)\n\n",
		  hopnr, addrNxt, nameNxt);
	  io.printline(PRESS, str);
	  return;
	}
      }
    }
    rtyp = response.val[1].value.intgr;
    rproto = response.val[2].value.intgr;
    metric = response.val[3].value.intgr;
    if (metric < 0 || metric > 128) metric = 0;
    switch (rtyp) {
      case 3: strcpy(typstr, "direct"); break;
      case 4: strcpy(typstr, "remote"); break;
      case 1: strcpy(typstr, "other");  break;
      default: strcpy(typstr, "invalid"); break;
    }
    switch (rproto) {
      case 1: strcpy(protstr, "other"); break;
      case 2: strcpy(protstr, "local"); break;
      case 3: strcpy(protstr, "netmgmt"); break;
      case 4: strcpy(protstr, "icmp"); break;
      case 5: strcpy(protstr, "egp"); break;
      case 6: strcpy(protstr, "ggp"); break;
      case 7: strcpy(protstr, "hello"); break;
      case 8: strcpy(protstr, "rip"); break;
      case 9: strcpy(protstr, "is-is"); break;
      case 10: strcpy(protstr, "es-is"); break;
      case 11: strcpy(protstr, "ciscoIgrp"); break;
      case 12: strcpy(protstr, "bbnSpfIgp"); break;
      case 13: strcpy(protstr, "oigp"); break;
      default: strcpy(protstr, "invalid"); break;
    }
    sprintf(str, "Hop %2d: %-16.16s %-30.30s %-7.7s  %-9.9s %3d\n", hopnr++,
	    addrNxt, nameNxt, typstr, protstr, metric);
    io.printline(NOPRESS, str);
    strcpy(addrprior, addrNxt);
    if (add && (g = lmgr.gtwy_find(addrNxt)) == NULLPTR) {
      sprintf(str, "%s %s %s", addrNxt, nameNxt, cidNxt);
      i = addcomponent(QUIETMODE, str, lmgr, io);
    }
    i = lmgr.saveNameAddr(addrNxt, addrNxt, nameNxt, cidNxt, YES, UNK); 
    strcpy(addrNxt, response.asciival[0]);      /* get the next hop */
    strcpy(nameNxt, "none");
    strcpy(aliasNxt, addrNxt);
#ifdef DEBUG
    sprintf(str, "    response: %s\n", addrNxt);
    io.printline(NOPRESS, str);
#endif
#ifdef DEBUG
	sprintf(str, "    rtyp: %d, rproto: %d\n", rtyp, rproto);
	io.printline(NOPRESS, str);
#endif
    addrNxtInWorld = lmgr.findNameAddr(addrNxt, aliasNxt, nameNxt, cid, 
				       snmp, typ);
    if (addrNxtInWorld == 1 || addrNxtInWorld == 2) strcpy(cidNxt, cid);
  }
  while ((strcmp(addrNxt, addrEnd) != 0) && (strcmp(addrNxt, addrprior) != 0) 
	 && (strcmp(addrNxt, "0.0.0.0") != 0));
  if (strcmp(addrNxt, "0.0.0.0") == 0) {
    sprintf(str, "Trace ended - remainder of route unknown.\n");
    io.printline(PRESS, str);
  }
  else {
    sprintf(str, "Hop %2d: %-16.16s %-30.30s\n", hopnr, addrEnd, nameEnd);
    io.printline(NOPRESS, str);
    if (add && (g = lmgr.gtwy_find(addrEnd)) == NULLPTR) {
      sprintf(str,"%s %s %s", addrEnd, nameEnd, cidEnd);
      i = addcomponent(QUIETMODE, str, lmgr, io);
    }
#ifdef DEBUG
    sprintf(str, "At end of trace, addrNxt = %s\n", addrNxt);
    io.printline(NOPRESS, str);
#endif
    io.printline(PRESS, "\n");
  }
}

#include "actmgr2.c"
